from dataclasses import dataclass
import enum
import glob
import os
from os import PathLike
import pathlib
import re
import shutil
import stat
from typing import Any, Final, Iterator, TypedDict

from ...config import GlobalConfig
from ...i18n import _
from ...log import RuyiLogger, humanize_list
from ...ruyipkg.atom import Atom
from ...ruyipkg.pkg_manifest import BoundPackageManifest, EmulatorProgDecl
from ...ruyipkg.profile import ProfileProxy
from ...utils.global_mode import ProvidesGlobalMode
from ...utils.l10n import match_lang_code
from ...utils.templating import render_template_str
from .emulator_cfg import ResolvedEmulatorProg


class ConfiguredTargetTuple(TypedDict):
    target: str
    toolchain_root: PathLike[Any]
    toolchain_sysroot: PathLike[Any] | None
    toolchain_flags: str
    binutils_flavor: str
    cc_flavor: str
    gcc_install_dir: PathLike[Any] | None


class ResolvedSysrootPkgSource(TypedDict):
    sysroot_dir: PathLike[Any]
    pkg_manifest: BoundPackageManifest
    gcc_install_dir: PathLike[Any] | None


class SysrootProvisionMode(enum.Enum):
    COPY_TREE = "copy-tree"
    SYMLINK_TREE = "symlink-tree"
    PROJECT_ROOTFS = "project-rootfs"


class VenvProvisionError(Exception):
    pass


PROJECTED_SYSROOT_ROOTS: Final = (
    "include",
    "lib",
    "lib32",
    "lib64",
    "usr/include",
    "usr/lib",
    "usr/lib32",
    "usr/lib64",
    "usr/libexec",
    "usr/share",
    "bin",
    "sbin",
)


@dataclass
class SysrootProjectionReport:
    copied_entries: int = 0
    skipped_entries: int = 0
    included_roots: int = 0


def _count_copytree_failures(exc: shutil.Error) -> int:
    if not exc.args:
        return 1

    failures = exc.args[0]
    if isinstance(failures, list):
        return len(failures)

    return 1


def _is_projected_rootfs_relpath(relpath: pathlib.PurePosixPath) -> bool:
    for root in PROJECTED_SYSROOT_ROOTS:
        projected_root = pathlib.PurePosixPath(root)
        if relpath == projected_root or projected_root in relpath.parents:
            return True

    return False


def _copy_projected_sysroot_symlink(
    source_root: pathlib.Path,
    dest_root: pathlib.Path,
    src: pathlib.Path,
    dest: pathlib.Path,
) -> None:
    target = os.readlink(src)
    if not target.startswith("/"):
        os.symlink(target, dest)
        return

    target_relpath = pathlib.PurePosixPath(target.removeprefix("/"))
    target_in_source = source_root.joinpath(*target_relpath.parts)
    if (
        _is_projected_rootfs_relpath(target_relpath)
        and (target_in_source.exists() or target_in_source.is_symlink())
    ):
        target_in_dest = dest_root.joinpath(*target_relpath.parts)
        os.symlink(os.path.relpath(target_in_dest, dest.parent), dest)
        return

    os.symlink(target, dest)


def _copy_projected_sysroot_entry(
    logger: RuyiLogger,
    source_root: pathlib.Path,
    dest_root: pathlib.Path,
    src: pathlib.Path,
    dest: pathlib.Path,
    report: SysrootProjectionReport,
) -> None:
    try:
        mode = src.lstat().st_mode
    except OSError as e:
        report.skipped_entries += 1
        logger.D(f"skipping sysroot entry {src}: {e}")
        return

    try:
        if stat.S_ISLNK(mode):
            dest.parent.mkdir(parents=True, exist_ok=True)
            _copy_projected_sysroot_symlink(source_root, dest_root, src, dest)
            report.copied_entries += 1
            return

        if stat.S_ISDIR(mode):
            dest.mkdir(parents=True, exist_ok=True)
            report.copied_entries += 1
            try:
                children = list(src.iterdir())
            except OSError as e:
                report.skipped_entries += 1
                logger.D(f"skipping sysroot directory contents {src}: {e}")
                return

            for child in children:
                _copy_projected_sysroot_entry(
                    logger,
                    source_root,
                    dest_root,
                    child,
                    dest / child.name,
                    report,
                )
            return

        if stat.S_ISREG(mode):
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src, dest, follow_symlinks=False)
            report.copied_entries += 1
            return
    except OSError as e:
        report.skipped_entries += 1
        logger.D(f"skipping sysroot entry {src}: {e}")
        return

    report.skipped_entries += 1
    logger.D(f"skipping unsupported sysroot entry {src}")


def _project_rootfs_sysroot(
    logger: RuyiLogger,
    src: pathlib.Path,
    dest: pathlib.Path,
) -> SysrootProjectionReport:
    report = SysrootProjectionReport()
    dest.mkdir(parents=True, exist_ok=False)

    for root in PROJECTED_SYSROOT_ROOTS:
        root_src = src / root
        if not root_src.exists() and not root_src.is_symlink():
            continue

        report.included_roots += 1
        _copy_projected_sysroot_entry(
            logger,
            src,
            dest,
            root_src,
            dest / root,
            report,
        )

    return report


def provision_sysroot(
    logger: RuyiLogger,
    src: pathlib.Path,
    dest: pathlib.Path,
    mode: SysrootProvisionMode,
    target_tuple: str,
) -> None:
    if mode == SysrootProvisionMode.SYMLINK_TREE:
        logger.D(f"symlinking sysroot for {target_tuple}")
        os.symlink(src, dest)
        return

    if mode == SysrootProvisionMode.COPY_TREE:
        logger.D(f"copying sysroot for {target_tuple}")
        try:
            shutil.copytree(
                src,
                dest,
                symlinks=True,
                ignore_dangling_symlinks=True,
            )
        except shutil.Error as e:
            failure_count = _count_copytree_failures(e)
            if failure_count == 1:
                reason = _("one entry could not be copied")
            else:
                reason = _("{count} entries could not be copied").format(
                    count=failure_count,
                )
            logger.F(
                _(
                    "cannot copy sysroot from [yellow]{src}[/]: {reason}"
                ).format(
                    src=src,
                    reason=reason,
                )
            )
            logger.I(
                _(
                    "Ruyi does not elevate privileges when creating virtual environments; use a sysroot readable by the current user, --symlink-sysroot-from-dir, or --project-sysroot-from-rootfs"
                )
            )
            logger.D(f"sysroot copy failure details: {e}")
            raise VenvProvisionError from e
        except OSError as e:
            logger.F(
                _("cannot copy sysroot from [yellow]{src}[/] to [green]{dest}[/]: {err}").format(
                    src=src,
                    dest=dest,
                    err=e,
                )
            )
            raise VenvProvisionError from e
        return

    if mode == SysrootProvisionMode.PROJECT_ROOTFS:
        logger.D(f"projecting rootfs sysroot for {target_tuple}")
        try:
            report = _project_rootfs_sysroot(logger, src, dest)
        except OSError as e:
            logger.F(
                _("cannot project sysroot from [yellow]{src}[/] to [green]{dest}[/]: {err}").format(
                    src=src,
                    dest=dest,
                    err=e,
                )
            )
            raise VenvProvisionError from e

        if report.included_roots == 0:
            logger.F(
                _(
                    "cannot project sysroot from [yellow]{src}[/]: no supported sysroot directories were found"
                ).format(src=src)
            )
            raise VenvProvisionError

        logger.I(
            _(
                "projected sysroot from [yellow]{src}[/]: copied {copied_count} entries, skipped {skipped_count} entries"
            ).format(
                src=src,
                copied_count=report.copied_entries,
                skipped_count=report.skipped_entries,
            )
        )
        if report.skipped_entries:
            logger.W(
                _(
                    "some unreadable or unsupported files were skipped; the projected sysroot may be incomplete"
                )
            )
        return

    raise NotImplementedError(mode)


def _resolve_sysroot_pkg_source(
    config: GlobalConfig,
    host: str,
    sysroot_atom_str: str,
) -> ResolvedSysrootPkgSource | int:
    """Resolve a sysroot package source from an atom string.

    Returns ResolvedSysrootPkgSource on success, or an error exit code (int).
    """
    logger = config.logger
    mr = config.repo

    gcc_pkg_atom = Atom.parse(sysroot_atom_str)
    gcc_pkg_pm = gcc_pkg_atom.match_in_repo(mr, config.include_prereleases)
    if gcc_pkg_pm is None:
        logger.F(
            _("cannot match a toolchain package with [yellow]{atom}[/]").format(
                atom=sysroot_atom_str,
            )
        )
        return 1

    if gcc_pkg_pm.toolchain_metadata is None:
        logger.F(
            _("the package [yellow]{atom}[/] is not a toolchain").format(
                atom=sysroot_atom_str,
            )
        )
        return 1

    gcc_pkg_root = config.lookup_binary_install_dir(
        host,
        gcc_pkg_pm.name_for_installation,
    )
    if gcc_pkg_root is None:
        logger.F(
            _("cannot find the installed directory for the sysroot package")
        )
        return 1

    tc_sysroot_relpath = gcc_pkg_pm.toolchain_metadata.included_sysroot
    if tc_sysroot_relpath is None:
        logger.F(
            _("sysroot is requested but the package [yellow]{atom}[/] does not contain one").format(
                atom=sysroot_atom_str,
            )
        )
        return 1

    sysroot_dir = pathlib.Path(gcc_pkg_root) / tc_sysroot_relpath

    # also figure the GCC include/libs path out for Clang to be able to
    # locate them
    gcc_install_dir = find_gcc_install_dir(
        gcc_pkg_root,
        # we should use the GCC-providing package's target tuple as that's
        # not guaranteed to be the same as llvm's
        gcc_pkg_pm.toolchain_metadata.target,
    )

    # for now, require this directory to be present (or clang would barely work)
    if gcc_install_dir is None:
        logger.F(
            _("cannot find a GCC include & lib directory in the sysroot package")
        )
        return 1

    return ResolvedSysrootPkgSource(
        sysroot_dir=sysroot_dir,
        pkg_manifest=gcc_pkg_pm,
        gcc_install_dir=gcc_install_dir,
    )


class VenvPackageInfo(TypedDict):
    repo_id: str
    category: str
    name: str
    version: str


class VenvMetadata(TypedDict):
    emulator_pkgs: dict[str, VenvPackageInfo]
    extra_pkgs: list[VenvPackageInfo]
    sysroot_pkg: VenvPackageInfo | None
    toolchain_pkgs: dict[str, VenvPackageInfo]


def _venv_pkg_info_from_pkg(pkg: BoundPackageManifest) -> VenvPackageInfo:
    return VenvPackageInfo(
        repo_id=pkg.repo_id,
        category=pkg.category,
        name=pkg.name,
        version=pkg.ver,
    )


def do_make_venv(
    config: GlobalConfig,
    host: str,
    profile_name: str,
    dest: pathlib.Path,
    with_sysroot: bool,
    override_name: str | None = None,
    tc_atoms_str: list[str] | None = None,
    emu_atom_str: str | None = None,
    sysroot_atom_str: str | None = None,
    copy_sysroot_dir_str: str | None = None,
    symlink_sysroot_dir_str: str | None = None,
    project_sysroot_dir_str: str | None = None,
    extra_cmd_atoms_str: list[str] | None = None,
) -> int:
    logger = config.logger

    explicit_sysroot_dir: pathlib.Path | None = None
    explicit_sysroot_gcc_install_dir: PathLike[Any] | None = None
    explicit_sysroot_pkg: VenvPackageInfo | None = None
    sysroot_provision_mode = SysrootProvisionMode.COPY_TREE

    if sysroot_atom_str is not None:
        result = _resolve_sysroot_pkg_source(config, host, sysroot_atom_str)
        if isinstance(result, int):
            return result

        explicit_sysroot_dir = pathlib.Path(result["sysroot_dir"])
        explicit_sysroot_gcc_install_dir = result["gcc_install_dir"]
        explicit_sysroot_pkg = _venv_pkg_info_from_pkg(result["pkg_manifest"])
    elif copy_sysroot_dir_str is not None:
        explicit_sysroot_dir = pathlib.Path(copy_sysroot_dir_str).resolve()
        if not explicit_sysroot_dir.is_dir():
            logger.F(
                _("the sysroot directory [yellow]{path}[/] does not exist").format(
                    path=copy_sysroot_dir_str,
                )
            )
            return 1
    elif symlink_sysroot_dir_str is not None:
        explicit_sysroot_dir = pathlib.Path(symlink_sysroot_dir_str).resolve()
        if not explicit_sysroot_dir.is_dir():
            logger.F(
                _("the sysroot directory [yellow]{path}[/] does not exist").format(
                    path=symlink_sysroot_dir_str,
                )
            )
            return 1
        sysroot_provision_mode = SysrootProvisionMode.SYMLINK_TREE
    elif project_sysroot_dir_str is not None:
        explicit_sysroot_dir = pathlib.Path(project_sysroot_dir_str).resolve()
        if not explicit_sysroot_dir.is_dir():
            logger.F(
                _("the rootfs directory [yellow]{path}[/] does not exist").format(
                    path=project_sysroot_dir_str,
                )
            )
            return 1
        sysroot_provision_mode = SysrootProvisionMode.PROJECT_ROOTFS

    # TODO: support omitting this if user only has one toolchain installed
    # this should come after implementation of local state cache
    if tc_atoms_str is None:
        logger.F(
            _("You have to specify at least one toolchain atom for now, e.g. [yellow]`-t gnu-plct`[/]")
        )
        return 1

    mr = config.repo

    profile = mr.get_profile(profile_name)
    if profile is None:
        logger.F(_("profile '{profile}' not found").format(profile=profile_name))
        return 1

    target_arch = ""
    seen_target_tuples: set[str] = set()
    targets: list[ConfiguredTargetTuple] = []
    warn_differing_target_arch = False
    venv_metadata = VenvMetadata(
        emulator_pkgs={},
        extra_pkgs=[],
        sysroot_pkg=None,
        toolchain_pkgs={},
    )

    for tc_atom_str in tc_atoms_str:
        tc_atom = Atom.parse(tc_atom_str)
        tc_pm = tc_atom.match_in_repo(mr, config.include_prereleases)
        if tc_pm is None:
            logger.F(_("cannot match a toolchain package with [yellow]{atom}[/]").format(
                atom=tc_atom_str,
            ))
            return 1

        if tc_pm.toolchain_metadata is None:
            logger.F(_("the package [yellow]{atom}[/] is not a toolchain").format(
                atom=tc_atom_str,
            ))
            return 1

        if not tc_pm.toolchain_metadata.satisfies_quirk_set(profile.need_quirks):
            logger.F(_(
                "the package [yellow]{atom}[/] does not support all necessary features for the profile [yellow]{profile}[/]"
                ).format(
                    atom=tc_atom_str,
                    profile=profile_name,
                    )
            )
            logger.I(
                _("quirks needed by profile:   {humanized_list}").format(
                humanized_list=humanize_list(profile.need_quirks, item_color='cyan'),
                )
            )
            logger.I(
                _("quirks provided by package: {humanized_list}").format(
                humanized_list=humanize_list(tc_pm.toolchain_metadata.quirks, item_color='yellow'),
                )
            )
            return 1

        target_tuple = tc_pm.toolchain_metadata.target
        if target_tuple in seen_target_tuples:
            logger.F(
                _("the target tuple [yellow]{target_tuple}[/] is already covered by one of the requested toolchains").format(
                    target_tuple=target_tuple,
                    )
            )
            logger.I(
                _("for now, only toolchains with differing target tuples can co-exist in one virtual environment")
            )
            return 1

        toolchain_root = config.lookup_binary_install_dir(
            host,
            tc_pm.name_for_installation,
        )
        if toolchain_root is None:
            logger.F(_("cannot find the installed directory for the toolchain"))
            return 1

        tc_sysroot_dir: PathLike[Any] | None = None
        gcc_install_dir: PathLike[Any] | None = None
        if with_sysroot:
            if explicit_sysroot_dir is not None:
                tc_sysroot_dir = explicit_sysroot_dir
                gcc_install_dir = explicit_sysroot_gcc_install_dir
                if explicit_sysroot_pkg is not None:
                    venv_metadata["sysroot_pkg"] = explicit_sysroot_pkg
            elif tc_sysroot_relpath := tc_pm.toolchain_metadata.included_sysroot:
                tc_sysroot_dir = pathlib.Path(toolchain_root) / tc_sysroot_relpath
                venv_metadata["sysroot_pkg"] = _venv_pkg_info_from_pkg(tc_pm)
            else:
                logger.F(
                    _("sysroot is requested but the toolchain package does not include one, and no explicit sysroot source is given")
                )
                return 1

        # derive flags for (the quirks of) this toolchain
        tc_flags = profile.get_common_flags(tc_pm.toolchain_metadata.quirks)

        # record the target tuple info to configure in the venv
        configured_target: ConfiguredTargetTuple = {
            "target": target_tuple,
            "toolchain_root": toolchain_root,
            "toolchain_sysroot": tc_sysroot_dir,
            "toolchain_flags": tc_flags,
            # assume clang is preferred if package contains clang
            # this is mostly true given most packages don't contain both
            "cc_flavor": "clang" if tc_pm.toolchain_metadata.has_clang else "gcc",
            # same for binutils provider flavor
            "binutils_flavor": (
                "llvm" if tc_pm.toolchain_metadata.has_llvm else "binutils"
            ),
            "gcc_install_dir": gcc_install_dir,
        }
        logger.D(f"configuration for {target_tuple}: {configured_target}")
        targets.append(configured_target)
        seen_target_tuples.add(target_tuple)
        venv_metadata["toolchain_pkgs"][target_tuple] = _venv_pkg_info_from_pkg(tc_pm)

        # record the target architecture for use in emulator package matching
        if not target_arch:
            target_arch = tc_pm.toolchain_metadata.target_arch
        elif target_arch != tc_pm.toolchain_metadata.target_arch:
            # first one wins
            warn_differing_target_arch = True

    if warn_differing_target_arch:
        logger.W(_("multiple toolchains specified with differing target architecture"))
        logger.I(
            _("using the target architecture of the first toolchain: [yellow]{arch}[/]").format(
                arch=target_arch,
                )
        )

    # Now handle the emulator.
    emu_progs = None
    emu_root: PathLike[Any] | None = None
    if emu_atom_str:
        emu_atom = Atom.parse(emu_atom_str)
        emu_pm = emu_atom.match_in_repo(mr, config.include_prereleases)
        if emu_pm is None:
            logger.F(_("cannot match an emulator package with [yellow]{atom}[/]").format(
                atom=emu_atom_str,
            ))
            return 1

        if emu_pm.emulator_metadata is None:
            logger.F(_("the package [yellow]{atom}[/] is not an emulator").format(
                atom=emu_atom_str,
            ))
            return 1

        emu_progs = list(emu_pm.emulator_metadata.list_for_arch(target_arch))
        if not emu_progs:
            logger.F(
                _("the emulator package [yellow]{atom}[/] does not support the target architecture [yellow]{arch}[/]").format(
                    atom=emu_atom_str,
                    arch=target_arch,
                )
            )
            return 1

        for prog in emu_progs:
            if not profile.check_emulator_flavor(
                prog.flavor,
                emu_pm.emulator_metadata.quirks,
            ):
                logger.F(
                    _("the package [yellow]{atom}[/] does not support all necessary features for the profile [yellow]{profile}[/]").format(
                        atom=emu_atom_str,
                        profile=profile_name,
                    )
                )
                logger.I(
                    _("quirks needed by profile:   {humanized_list}").format(
                        humanized_list=humanize_list(profile.get_needed_emulator_pkg_flavors(prog.flavor), item_color='cyan'),
                    )
                )
                logger.I(
                    _("quirks provided by package: {humanized_list}").format(
                        humanized_list=humanize_list(emu_pm.emulator_metadata.quirks or [], item_color='yellow'),
                    )
                )
                return 1

        emu_root = config.lookup_binary_install_dir(
            host,
            emu_pm.name_for_installation,
        )
        if emu_root is None:
            logger.F(_("cannot find the installed directory for the emulator"))
            return 1

        venv_metadata["emulator_pkgs"][target_arch] = _venv_pkg_info_from_pkg(emu_pm)

    # Now resolve extra commands to provide in the venv.
    extra_cmds: dict[str, str] = {}
    if extra_cmd_atoms_str:
        for extra_cmd_atom_str in extra_cmd_atoms_str:
            extra_cmd_atom = Atom.parse(extra_cmd_atom_str)
            extra_cmd_pm = extra_cmd_atom.match_in_repo(
                mr,
                config.include_prereleases,
            )
            if extra_cmd_pm is None:
                logger.F(
                    _("cannot match an extra command package with [yellow]{atom}[/]").format(
                        atom=extra_cmd_atom_str,
                    )
                )
                return 1

            extra_cmd_bm = extra_cmd_pm.binary_metadata
            if not extra_cmd_bm:
                logger.F(
                    _("the package [yellow]{atom}[/] is not a binary-providing package").format(
                        atom=extra_cmd_atom_str,
                    )
                )
                return 1

            extra_cmds_decl = extra_cmd_bm.get_commands_for_host(host)
            if not extra_cmds_decl:
                logger.W(
                    _("the package [yellow]{atom}[/] does not provide any command for host [yellow]{host}[/], ignoring").format(
                        atom=extra_cmd_atom_str,
                        host=host,
                    )
                )
                continue

            cmd_root = config.lookup_binary_install_dir(
                host,
                extra_cmd_pm.name_for_installation,
            )
            if cmd_root is None:
                logger.F(
                    _("cannot find the installed directory for the package [yellow]{pkg}[/]").format(
                        pkg=extra_cmd_pm.name_for_installation,
                    )
                )
                return 1
            cmd_root = pathlib.Path(cmd_root)

            venv_metadata["extra_pkgs"].append(_venv_pkg_info_from_pkg(extra_cmd_pm))

            for cmd, cmd_rel_path in extra_cmds_decl.items():
                # resolve the command path
                cmd_path = (cmd_root / cmd_rel_path).resolve()
                if not cmd_path.is_relative_to(cmd_root):
                    # we don't allow commands to resolve outside of the
                    # providing package's install root
                    logger.F(
                        _("internal error: resolved command path is outside of the providing package")
                    )
                    return 1

                # add the command to the list
                extra_cmds[cmd] = str(cmd_path)

    if override_name is not None:
        logger.I(
            _("Creating a Ruyi virtual environment [cyan]'{name}'[/] at [green]{dest}[/]...").format(
                name=override_name,
                dest=dest,
            )
        )
    else:
        logger.I(
            _("Creating a Ruyi virtual environment at [green]{dest}[/]...").format(
                dest=dest,
            )
        )

    maker = VenvMaker(
        config,
        profile,
        targets,
        dest.resolve(),
        emu_progs,
        emu_root,
        extra_cmds,
        venv_metadata,
        override_name,
        sysroot_provision_mode,
    )
    try:
        maker.provision()
    except VenvProvisionError:
        return 1

    # TODO: move the template to PO
    locale = match_lang_code(config.lang_code, avail=("en", "zh_CN"))
    logger.I(
        render_template_str(
            f"prompt.venv-created.{locale}.txt",
            {
                "sysroot": maker.sysroot_destdir(None),
            },
        )
    )

    return 0


def find_gcc_install_dir(
    install_root: PathLike[Any],
    target_tuple: str,
) -> PathLike[Any] | None:
    # check $PREFIX/lib/gcc/$TARGET/*
    search_root = pathlib.Path(install_root) / "lib" / "gcc" / target_tuple
    try:
        for p in search_root.iterdir():
            # only want the first one (should be the only one)
            return p
    except FileNotFoundError:
        pass

    # nothing?
    return None


class VenvMaker:
    """Performs the actual creation of a Ruyi virtual environment."""

    def __init__(
        self,
        gc: GlobalConfig,
        profile: ProfileProxy,
        targets: list[ConfiguredTargetTuple],
        dest: PathLike[Any],
        emulator_progs: list[EmulatorProgDecl] | None,
        emulator_root: PathLike[Any] | None,
        extra_cmds: dict[str, str] | None,
        metadata: VenvMetadata,
        override_name: str | None = None,
        sysroot_provision_mode: SysrootProvisionMode = SysrootProvisionMode.COPY_TREE,
    ) -> None:
        self.gc = gc
        self.profile = profile
        self.targets = targets
        self.venv_root = pathlib.Path(dest)
        self.emulator_progs = emulator_progs
        self.emulator_root = emulator_root
        self.extra_cmds = extra_cmds or {}
        self.metadata = metadata
        self.override_name = override_name
        self.sysroot_provision_mode = sysroot_provision_mode

        self.bindir = self.venv_root / "bin"

    @property
    def logger(self) -> RuyiLogger:
        return self.gc.logger

    def render_and_write(
        self,
        dest: PathLike[Any],
        template_name: str,
        data: dict[str, Any],
    ) -> None:
        self.logger.D(f"rendering template '{template_name}' with data {data}")
        content = render_template_str(template_name, data).encode("utf-8")
        self.logger.D(f"writing {dest}")
        with open(dest, "wb") as fp:
            fp.write(content)

    def sysroot_srcdir(self, target_tuple: str | None) -> pathlib.Path | None:
        if target_tuple is None:
            # check the primary target
            if s := self.targets[0]["toolchain_sysroot"]:
                return pathlib.Path(s)
            return None

        # check if we have this target
        for t in self.targets:
            if t["target"] != target_tuple:
                continue
            if s := t["toolchain_sysroot"]:
                return pathlib.Path(s)

        return None

    def has_sysroot_for(self, target_tuple: str | None) -> bool:
        return self.sysroot_srcdir(target_tuple) is not None

    def sysroot_destdir(self, target_tuple: str | None) -> pathlib.Path | None:
        if not self.has_sysroot_for(target_tuple):
            return None

        dirname = f"sysroot.{target_tuple}" if target_tuple is not None else "sysroot"
        return self.venv_root / dirname

    def provision(self) -> None:
        venv_root = self.venv_root
        bindir = self.bindir

        venv_root.mkdir()
        bindir.mkdir()

        env_data = {
            "profile": self.profile.id,
            "sysroot": self.sysroot_destdir(None),
            "metadata": self.metadata,
        }
        self.render_and_write(
            venv_root / "ruyi-venv.toml",
            "ruyi-venv.toml",
            env_data,
        )

        for i, tgt in enumerate(self.targets):
            is_primary = i == 0
            self.provision_target(tgt, is_primary)

        if self.extra_cmds:
            symlink_binaries(
                self.gc,
                self.logger,
                bindir,
                src_cmds_names=list(self.extra_cmds.keys()),
            )

        template_data = {
            "RUYI_VENV": str(venv_root),
            "RUYI_VENV_NAME": self.override_name,
        }

        self.render_and_write(
            bindir / "ruyi-activate",
            "ruyi-activate.bash",
            template_data,
        )

        self.render_and_write(
            bindir / "ruyi-activate.fish",
            "ruyi-activate.fish",
            template_data,
        )

        qemu_bin: PathLike[Any] | None = None
        profile_emu_env: dict[str, str] | None = None
        if self.emulator_root is not None and self.emulator_progs:
            resolved_emu_progs = [
                ResolvedEmulatorProg.new(
                    p,
                    self.emulator_root,
                    self.profile,
                    self.sysroot_destdir(None),
                )
                for p in self.emulator_progs
            ]
            binfmt_data = {
                "resolved_progs": resolved_emu_progs,
            }
            self.render_and_write(
                venv_root / "binfmt.conf",
                "binfmt.conf",
                binfmt_data,
            )

            for i, p in enumerate(self.emulator_progs):
                if not p.is_qemu:
                    continue

                qemu_bin = pathlib.Path(self.emulator_root) / p.relative_path
                profile_emu_env = resolved_emu_progs[i].env

                self.logger.D("symlinking the ruyi-qemu wrapper")
                os.symlink(self.gc.self_exe, bindir / "ruyi-qemu")

        # provide initial cached configuration to venv
        self.render_and_write(
            venv_root / "ruyi-cache.v2.toml",
            "ruyi-cache.toml",
            self.make_venv_cache_data(
                qemu_bin,
                self.extra_cmds,
                profile_emu_env,
            ),
        )

    def make_venv_cache_data(
        self,
        qemu_bin: PathLike[Any] | None,
        extra_cmds: dict[str, str],
        profile_emu_env: dict[str, str] | None,
    ) -> dict[str, object]:
        targets_cache_data: dict[str, object] = {
            tgt["target"]: {
                "toolchain_bindir": str(pathlib.Path(tgt["toolchain_root"]) / "bin"),
                "toolchain_sysroot": self.sysroot_destdir(tgt["target"]),
                "toolchain_flags": tgt["toolchain_flags"],
                "gcc_install_dir": tgt["gcc_install_dir"],
            }
            for tgt in self.targets
        }

        cmd_metadata_map = make_cmd_metadata_map(self.logger, self.targets)

        # add extra cmds that are not associated with any target
        for cmd, dest in extra_cmds.items():
            if cmd in cmd_metadata_map:
                self.logger.W(
                    _("extra command {cmd} is already provided by another package, overriding it").format(
                        cmd=cmd,
                        )
                )
            cmd_metadata_map[cmd] = {
                "dest": dest,
                "target_tuple": "",
            }

        return {
            "profile_emu_env": profile_emu_env,
            "qemu_bin": qemu_bin,
            "targets": targets_cache_data,
            "cmd_metadata_map": cmd_metadata_map,
        }

    def provision_target(
        self,
        tgt: ConfiguredTargetTuple,
        is_primary: bool,
    ) -> None:
        venv_root = self.venv_root
        bindir = self.bindir
        target_tuple = tgt["target"]

        # getting the destdir this way ensures it's suffixed with the target
        # tuple
        if sysroot_destdir := self.sysroot_destdir(target_tuple):
            sysroot_srcdir = tgt["toolchain_sysroot"]
            assert sysroot_srcdir is not None
            sysroot_srcdir = pathlib.Path(sysroot_srcdir)

            provision_sysroot(
                self.logger,
                sysroot_srcdir,
                sysroot_destdir,
                self.sysroot_provision_mode,
                target_tuple,
            )

            if is_primary:
                self.logger.D("symlinking primary sysroot into place")
                primary_sysroot_destdir = self.sysroot_destdir(None)
                assert primary_sysroot_destdir is not None
                os.symlink(sysroot_destdir.name, primary_sysroot_destdir)

        self.logger.D(f"symlinking {target_tuple} binaries into venv")
        toolchain_bindir = pathlib.Path(tgt["toolchain_root"]) / "bin"
        symlink_binaries(self.gc, self.logger, bindir, src_bindir=toolchain_bindir)

        make_llvm_tool_aliases(
            self.logger,
            bindir,
            target_tuple,
            tgt["binutils_flavor"] == "llvm",
            tgt["cc_flavor"] == "clang",
        )

        # CMake toolchain file & Meson cross file
        if tgt["cc_flavor"] == "clang":
            cc_path = bindir / "clang"
            cxx_path = bindir / "clang++"
        elif tgt["cc_flavor"] == "gcc":
            cc_path = bindir / f"{target_tuple}-gcc"
            cxx_path = bindir / f"{target_tuple}-g++"
        else:
            raise NotImplementedError

        if tgt["binutils_flavor"] == "binutils":
            meson_additional_binaries = {
                "ar": bindir / f"{target_tuple}-ar",
                "nm": bindir / f"{target_tuple}-nm",
                "objcopy": bindir / f"{target_tuple}-objcopy",
                "objdump": bindir / f"{target_tuple}-objdump",
                "ranlib": bindir / f"{target_tuple}-ranlib",
                "readelf": bindir / f"{target_tuple}-readelf",
                "strip": bindir / f"{target_tuple}-strip",
            }
        elif tgt["binutils_flavor"] == "llvm":
            meson_additional_binaries = {
                "ar": bindir / "llvm-ar",
                "nm": bindir / "llvm-nm",
                "objcopy": bindir / "llvm-objcopy",
                "objdump": bindir / "llvm-objdump",
                "ranlib": bindir / "llvm-ranlib",
                "readelf": bindir / "llvm-readelf",
                "strip": bindir / "llvm-strip",
            }
        else:
            raise NotImplementedError

        cmake_toolchain_file_path = venv_root / f"toolchain.{target_tuple}.cmake"
        toolchain_file_data = {
            "cc": cc_path,
            "cxx": cxx_path,
            "processor": self.profile.arch,
            "sysroot": self.sysroot_destdir(target_tuple),
            "venv_root": venv_root,
            "cmake_toolchain_file": str(cmake_toolchain_file_path),
            "meson_additional_binaries": meson_additional_binaries,
        }
        self.render_and_write(
            cmake_toolchain_file_path,
            "toolchain.cmake",
            toolchain_file_data,
        )

        meson_cross_file_path = venv_root / f"meson-cross.{target_tuple}.ini"
        self.render_and_write(
            meson_cross_file_path,
            "meson-cross.ini",
            toolchain_file_data,
        )

        if is_primary:
            self.logger.D(
                f"making cmake & meson file symlinks to primary target {target_tuple}"
            )
            primary_cmake_toolchain_file_path = venv_root / "toolchain.cmake"
            primary_meson_cross_file_path = venv_root / "meson-cross.ini"
            os.symlink(
                cmake_toolchain_file_path.name,
                primary_cmake_toolchain_file_path,
            )
            os.symlink(meson_cross_file_path.name, primary_meson_cross_file_path)


def iter_binaries_to_symlink(
    logger: RuyiLogger,
    bindir: pathlib.Path,
) -> Iterator[pathlib.Path]:
    for filename in glob.iglob("*", root_dir=bindir):
        src_cmd_path = bindir / filename
        if not is_executable(src_cmd_path):
            logger.D(f"skipping non-executable {filename} in src bindir")
            continue

        if should_ignore_symlinking(filename):
            logger.D(f"skipping command {filename} explicitly")
            continue

        yield bindir / filename


class CmdMetadataEntry(TypedDict):
    dest: str
    target_tuple: str


def make_cmd_metadata_map(
    logger: RuyiLogger,
    targets: list[ConfiguredTargetTuple],
) -> dict[str, CmdMetadataEntry]:
    result: dict[str, CmdMetadataEntry] = {}
    for tgt in targets:
        # TODO: dedup this and provision_target
        toolchain_bindir = pathlib.Path(tgt["toolchain_root"]) / "bin"
        for cmd in iter_binaries_to_symlink(logger, toolchain_bindir):
            result[cmd.name] = {
                "dest": str(cmd),
                "target_tuple": tgt["target"],
            }
    return result


def symlink_binaries(
    gm: ProvidesGlobalMode,
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    *,
    src_bindir: PathLike[Any] | None = None,
    src_cmds_names: list[str] | None = None,
) -> None:
    dest_binpath = pathlib.Path(dest_bindir)
    self_exe_path = gm.self_exe

    if src_bindir is not None:
        src_binpath = pathlib.Path(src_bindir)
        for src_cmd_path in iter_binaries_to_symlink(logger, src_binpath):
            filename = src_cmd_path.name

            # symlink self to dest with the name of this command
            dest_path = dest_binpath / filename
            logger.D(f"making ruyi symlink to {self_exe_path} at {dest_path}")
            os.symlink(self_exe_path, dest_path)
        return

    if src_cmds_names is not None:
        for cmd in src_cmds_names:
            # symlink self to dest with the name of this command
            dest_path = dest_binpath / cmd
            logger.D(f"making ruyi symlink to {self_exe_path} at {dest_path}")
            os.symlink(self_exe_path, dest_path)
        return

    raise ValueError(
        "internal error: either src_bindir or src_cmds_names must be provided"
    )


LLVM_BINUTILS_ALIASES: Final = {
    "addr2line": "llvm-addr2line",
    "ar": "llvm-ar",
    "as": "llvm-as",
    "c++filt": "llvm-cxxfilt",
    "gcc-ar": "llvm-ar",
    "gcc-nm": "llvm-nm",
    "gcc-ranlib": "llvm-ranlib",
    # 'gcov': 'llvm-cov',  # I'm not sure if this is correct
    "ld": "ld.lld",
    "nm": "llvm-nm",
    "objcopy": "llvm-objcopy",
    "objdump": "llvm-objdump",
    "ranlib": "llvm-ranlib",
    "readelf": "llvm-readelf",
    "size": "llvm-size",
    "strings": "llvm-strings",
    "strip": "llvm-strip",
}

CLANG_GCC_ALIASES: Final = {
    "c++": "clang++",
    "cc": "clang",
    "cpp": "clang-cpp",
    "g++": "clang++",
    "gcc": "clang",
}


def make_llvm_tool_aliases(
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    target_tuple: str,
    do_binutils: bool,
    do_clang: bool,
) -> None:
    if do_binutils:
        make_compat_symlinks(logger, dest_bindir, target_tuple, LLVM_BINUTILS_ALIASES)
    if do_clang:
        make_compat_symlinks(logger, dest_bindir, target_tuple, CLANG_GCC_ALIASES)


def make_compat_symlinks(
    logger: RuyiLogger,
    dest_bindir: PathLike[Any],
    target_tuple: str,
    aliases: dict[str, str],
) -> None:
    destdir = pathlib.Path(dest_bindir)
    for compat_basename, symlink_target in aliases.items():
        compat_name = f"{target_tuple}-{compat_basename}"
        logger.D(f"making compat symlink: {compat_name} -> {symlink_target}")
        os.symlink(symlink_target, destdir / compat_name)


def is_executable(p: PathLike[Any]) -> bool:
    return os.access(p, os.F_OK | os.X_OK)


def should_ignore_symlinking(c: str) -> bool:
    return is_command_specific_to_ct_ng(c) or is_command_versioned_cc(c)


def is_command_specific_to_ct_ng(c: str) -> bool:
    return c.endswith("populate") or c.endswith("ct-ng.config")


VERSIONED_CC_RE: Final = re.compile(
    r"(?:^|-)(?:g?cc|c\+\+|g\+\+|cpp|clang|clang\+\+)-[0-9.]+$"
)


def is_command_versioned_cc(c: str) -> bool:
    return VERSIONED_CC_RE.search(c) is not None



from os import PathLike
from typing import Any, Sequence

from ..i18n import _


class InvalidConfigSectionError(Exception):
    def __init__(self, section: str) -> None:
        super().__init__()
        self._section = section

    def __str__(self) -> str:
        return _("invalid config section: {section}").format(section=self._section)

    def __repr__(self) -> str:
        return f"InvalidConfigSectionError({self._section!r})"


class InvalidConfigKeyError(Exception):
    def __init__(self, key: str | Sequence[str]) -> None:
        super().__init__()
        self._key = key

    def __str__(self) -> str:
        return _("invalid config key: {key}").format(key=self._key)

    def __repr__(self) -> str:
        return f"InvalidConfigKeyError({self._key:!r})"


class InvalidConfigValueTypeError(TypeError):
    def __init__(
        self,
        key: str | Sequence[str],
        val: object | None,
        expected: type | Sequence[type],
    ) -> None:
        super().__init__()
        self._key = key
        self._val = val
        self._expected = expected

    def __str__(self) -> str:
        return _(
            "invalid value type for config key {key}: {actual_type}, expected {expected_type}"
        ).format(
            key=self._key,
            actual_type=type(self._val),
            expected_type=self._expected,
        )

    def __repr__(self) -> str:
        return f"InvalidConfigValueTypeError({self._key!r}, {self._val!r}, {self._expected:!r})"


class InvalidConfigValueError(ValueError):
    def __init__(
        self,
        key: str | Sequence[str] | None,
        val: object | None,
        typ: type | Sequence[type],
    ) -> None:
        super().__init__()
        self._key = key
        self._val = val
        self._typ = typ

    def __str__(self) -> str:
        return _("invalid config value for key {key} (type {typ}): {val}").format(
            key=self._key,
            typ=self._typ,
            val=self._val,
        )

    def __repr__(self) -> str:
        return (
            f"InvalidConfigValueError({self._key:!r}, {self._val:!r}, {self._typ:!r})"
        )


class MalformedConfigFileError(Exception):
    def __init__(self, path: PathLike[Any]) -> None:
        super().__init__()
        self._path = path

    def __str__(self) -> str:
        return _("malformed config file: {path}").format(path=self._path)

    def __repr__(self) -> str:
        return f"MalformedConfigFileError({self._path:!r})"


class ProtectedGlobalConfigError(Exception):
    def __init__(self, key: str | Sequence[str]) -> None:
        super().__init__()
        self._key = key

    def __str__(self) -> str:
        return _("attempt to modify protected global config key: {key}").format(
            key=self._key,
        )

    def __repr__(self) -> str:
        return f"ProtectedGlobalConfigError({self._key!r})"



"""Tests for [[repos]] config parsing and validation."""

from typing import TYPE_CHECKING

from ruyi.config import GlobalConfig

if TYPE_CHECKING:
    from tests.fixtures import MockGlobalModeProvider
    from ruyi.log import RuyiLogger


class TestReposConfigParsing:
    def test_no_repos_section(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        # Only the default entry should be present.
        assert len(gc.repo_entries) == 1
        assert gc.repo_entries[0].id == "ruyisdk"

    def test_single_extra_repo(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "my-vendor",
                        "remote": "https://example.invalid/repo.git",
                        "priority": 50,
                    }
                ]
            },
            is_global_scope=False,
        )
        # Clear cached property to pick up the new entries.
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        assert len(gc.repo_entries) == 2
        ids = [e.id for e in gc.repo_entries]
        assert "ruyisdk" in ids
        assert "my-vendor" in ids

    def test_repos_sorted_by_priority(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "high-prio",
                        "remote": "https://example.invalid/high.git",
                        "priority": 100,
                    },
                    {
                        "id": "low-prio",
                        "remote": "https://example.invalid/low.git",
                        "priority": -10,
                    },
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        entries = gc.repo_entries
        assert len(entries) == 3
        # Should be sorted by priority ascending.
        priorities = [e.priority for e in entries]
        assert priorities == sorted(priorities)

    def test_rejects_invalid_id(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "INVALID-CAPS",
                        "remote": "https://example.invalid/repo.git",
                    },
                    {
                        "id": "",
                        "remote": "https://example.invalid/repo.git",
                    },
                    {
                        "id": "-starts-with-dash",
                        "remote": "https://example.invalid/repo.git",
                    },
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        # None of the invalid entries should be added.
        assert len(gc.repo_entries) == 1
        assert gc.repo_entries[0].id == "ruyisdk"

    def test_rejects_reserved_id(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "ruyisdk",
                        "remote": "https://example.invalid/repo.git",
                        "priority": 999,
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        # Should still have only the default entry.
        assert len(gc.repo_entries) == 1
        assert gc.repo_entries[0].id == "ruyisdk"
        assert gc.repo_entries[0].priority == 0  # unmodified

    def test_rejects_duplicate_ids(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "vendor-a",
                        "remote": "https://example.invalid/a.git",
                        "priority": 10,
                    },
                    {
                        "id": "vendor-a",
                        "remote": "https://example.invalid/a-dup.git",
                        "priority": 20,
                    },
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        vendor_entries = [e for e in gc.repo_entries if e.id == "vendor-a"]
        assert len(vendor_entries) == 1
        assert vendor_entries[0].priority == 10  # first one wins

    def test_rejects_no_remote_no_local(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "broken",
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        assert len(gc.repo_entries) == 1
        assert gc.repo_entries[0].id == "ruyisdk"

    def test_rejects_relative_local_path(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "bad-path",
                        "local": "relative/path",
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        assert len(gc.repo_entries) == 1

    def test_local_only_repo(
        self,
        tmp_path: "object",
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "local-only",
                        "local": "/tmp/some/repo",
                        "priority": 5,
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        entries = gc.repo_entries
        local_entry = [e for e in entries if e.id == "local-only"]
        assert len(local_entry) == 1
        assert local_entry[0].local_path == "/tmp/some/repo"
        assert local_entry[0].remote == ""
        assert local_entry[0].active is True

    def test_defaults_for_optional_fields(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "minimal",
                        "remote": "https://example.invalid/repo.git",
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        entry = [e for e in gc.repo_entries if e.id == "minimal"][0]
        assert entry.name == "minimal"  # defaults to id
        assert entry.branch == "main"
        assert entry.priority == 0
        assert entry.active is True
        assert entry.local_path is None

    def test_inactive_repo(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc._apply_config(
            {
                "repos": [
                    {
                        "id": "disabled",
                        "remote": "https://example.invalid/repo.git",
                        "active": False,
                    }
                ]
            },
            is_global_scope=False,
        )
        if "repo_entries" in gc.__dict__:
            del gc.__dict__["repo_entries"]

        entry = [e for e in gc.repo_entries if e.id == "disabled"][0]
        assert entry.active is False



from typing import Final

from .i18n import d_

RUYI_SEMVER: Final = "0.49.0"
RUYI_USER_AGENT: Final = f"ruyi/{RUYI_SEMVER}"

COPYRIGHT_NOTICE: Final = d_(
    """\
Copyright (C) Institute of Software, Chinese Academy of Sciences (ISCAS).
All rights reserved.
License: Apache-2.0 <https://www.apache.org/licenses/LICENSE-2.0>
\
"""
)

MPL_REDIST_NOTICE: Final = d_(
    """\
This distribution of ruyi contains code licensed under the Mozilla Public
License 2.0 (https://mozilla.org/MPL/2.0/). You can get the respective
project's sources from the project's official website:

* certifi: https://github.com/certifi/python-certifi
\
"""
)



import atexit
import os
import sys
from typing import Final, TYPE_CHECKING

from ..config import GlobalConfig
from ..i18n import _
from ..telemetry.scope import TelemetryScope
from ..utils.global_mode import (
    GlobalModeProvider,
    is_cli_completion_script_requested,
)
from ..version import RUYI_SEMVER
from . import RUYI_ENTRYPOINT_NAME
from .oobe import OOBE

ALLOWED_RUYI_ENTRYPOINT_NAMES: Final = (
    RUYI_ENTRYPOINT_NAME,
    f"{RUYI_ENTRYPOINT_NAME}.exe",
    f"{RUYI_ENTRYPOINT_NAME}.bin",  # Nuitka one-file program cache
    "__main__.py",
)
VERSION_QUERY_FLAGS: Final = frozenset(("-V", "--version"))
VERSION_QUERY_SUBCOMMAND: Final = "version"


def is_called_as_ruyi(argv0: str) -> bool:
    return os.path.basename(argv0).lower() in ALLOWED_RUYI_ENTRYPOINT_NAMES


def should_prompt_for_renaming(argv0: str) -> bool:
    # We need to allow things like "ruyi-qemu" through, to not break our mux.
    # Only consider filenames starting with both our name *and* version to be
    # un-renamed onefile artifacts that warrant a rename prompt.
    likely_artifact_name_prefix = f"{RUYI_ENTRYPOINT_NAME}-{RUYI_SEMVER}."
    return os.path.basename(argv0).lower().startswith(likely_artifact_name_prefix)


def is_version_query(argv: list[str]) -> bool:
    # Scan flags broadly for issue #453, while only accepting the subcommand
    # spelling in command position so positional values named "version" do not
    # suppress OOBE/telemetry for unrelated commands.
    for arg in argv[1:]:
        if arg in VERSION_QUERY_FLAGS:
            return True

    for arg in argv[1:]:
        if arg == "--porcelain":
            continue
        return arg == VERSION_QUERY_SUBCOMMAND

    return False


def main(gm: GlobalModeProvider, gc: GlobalConfig, argv: list[str]) -> int:
    logger = gc.logger
    is_completion_script_invocation = is_cli_completion_script_requested(argv)
    is_ruyi_invocation = is_called_as_ruyi(gm.argv0)
    # Version queries must be side-effect-free: issue #453 showed that OOBE
    # could otherwise prompt before printing the version and consume first-run
    # telemetry state.
    skip_telemetry = is_ruyi_invocation and is_version_query(argv)

    # do not init telemetry or OOBE on shell completion invocations, because
    # our output isn't meant for humans in that case, and a "real" invocation
    # will likely follow shortly after
    if (
        not gm.is_cli_autocomplete
        and not is_completion_script_invocation
        and not skip_telemetry
    ):
        oobe = OOBE(gc)

        tm = gc.telemetry
        tm.check_first_run_status()
        tm.init_installation(False)
        atexit.register(tm.flush)
        oobe.handlers.append(tm.oobe_prompt)

        oobe.maybe_prompt()

    if not is_ruyi_invocation:
        if should_prompt_for_renaming(gm.argv0):
            logger.F(
                _(
                    "the {ruyi_exe} executable must be named [green]'{expected_name}'[/] to work"
                ).format(
                    ruyi_exe=RUYI_ENTRYPOINT_NAME,
                    expected_name=RUYI_ENTRYPOINT_NAME,
                )
            )
            logger.I(
                _("it is now [yellow]'{current_name}'[/]").format(
                    current_name=gm.argv0,
                )
            )
            logger.I(_("please rename the command file and retry"))
            return 1

        from ..mux.runtime import mux_main

        # record an invocation and the command name being proxied to
        gc.telemetry.record(
            TelemetryScope(None),
            "cli:mux-invocation-v1",
            target=os.path.basename(gm.argv0),
        )

        return mux_main(gm, gc, argv)

    import ruyi
    from .cmd import RootCommand
    from . import builtin_commands

    del builtin_commands

    if TYPE_CHECKING:
        from .cmd import CLIEntrypoint

    p = RootCommand.build_argparse(gc)

    # We have to ensure argcomplete is only requested when it's supposed to,
    # as the argcomplete import is very costly in terms of startup time, and
    # that the package name completer requires the whole repo to be synced
    # (which may not be the case for an out-of-the-box experience).
    if gm.is_cli_autocomplete:
        import argcomplete
        from .completer import NoneCompleter

        argcomplete.autocomplete(
            p,
            always_complete_options=True,
            default_completer=NoneCompleter(),
        )

    args = p.parse_args(argv[1:])
    # for getting access to the argparse parser in the CLI entrypoint
    args._parser = p  # pylint: disable=protected-access

    gm.is_porcelain = args.porcelain

    nuitka_info = "not compiled"
    if hasattr(ruyi, "__compiled__"):
        nuitka_info = f"__compiled__ = {ruyi.__compiled__}"

    logger.D(
        f"__main__.__file__ = {gm.main_file}, sys.executable = {sys.executable}, {nuitka_info}"
    )
    logger.D(f"argv[0] = {gm.argv0}, self_exe = {gm.self_exe}")
    logger.D(f"args={args}")

    func: "CLIEntrypoint" = args.func

    # record every invocation's subcommand for better insight into usage
    # frequencies
    try:
        telemetry_key: str = args.tele_key
    except AttributeError:
        logger.F(_("internal error: CLI entrypoint was added without a telemetry key"))
        return 1

    # Special-case the `--output-completion-script` argument; treat it as if
    # "ruyi completion-script" were called.
    completion_script: str | None = getattr(args, "completion_script", None)
    if is_completion_script_invocation and completion_script is not None:
        return func(gc, args)

    if not skip_telemetry:
        tm = gc.telemetry
        tm.print_telemetry_notice()

        # Do not record `ruyi telemetry --cron-upload` invocations.
        skip_recording_invocation = telemetry_key == "telemetry" and getattr(
            args,
            "cron_upload",
            False,
        )
        if not skip_recording_invocation:
            tm.record(
                TelemetryScope(None),
                "cli:invocation-v1",
                key=telemetry_key,
            )

    return func(gc, args)



import pathlib
from typing import TYPE_CHECKING

import pygit2

from ruyi.ruyipkg.composite_repo import CompositeRepo
from ruyi.ruyipkg.repo import RepoEntry

if TYPE_CHECKING:
    from tests.fixtures import MockGlobalModeProvider
    from ruyi.log import RuyiLogger


def _make_entry(
    repo_id: str = "test-repo",
    priority: int = 0,
    active: bool = True,
    local_path: str | None = None,
) -> RepoEntry:
    return RepoEntry(
        id=repo_id,
        name=f"Test Repo {repo_id}",
        remote="https://example.invalid/repo.git",
        branch="main",
        local_path=local_path,
        priority=priority,
        active=active,
    )


def _init_repo_dir(root: pathlib.Path) -> None:
    """Initialize a bare-minimum git repo at root for MetadataRepo."""
    root.mkdir(parents=True, exist_ok=True)
    pygit2.init_repository(str(root))


def _write_manifest(
    root: pathlib.Path,
    category: str,
    name: str,
    ver: str,
    *,
    slug: str | None = None,
    desc: str = "test package",
) -> None:
    """Write a minimal package manifest TOML file into a repo tree."""
    pkg_dir = root / "packages" / category / name
    pkg_dir.mkdir(parents=True, exist_ok=True)

    lines = [
        'format = "v1"',
        'kind = ["source"]',
        "distfiles = []",
        "",
        "[metadata]",
        f'desc = "{desc}"',
    ]
    if slug is not None:
        lines.append(f'slug = "{slug}"')
    lines.extend(
        [
            "",
            "[metadata.vendor]",
            'name = "test-vendor"',
        ]
    )
    (pkg_dir / f"{ver}.toml").write_text("\n".join(lines) + "\n")


def _write_profile_plugin(
    root: pathlib.Path,
    arch: str,
    profile_ids: list[str],
    *,
    quirks_by_profile: dict[str, list[str]] | None = None,
) -> None:
    plugin_dir = root / "plugins" / f"ruyi-profile-{arch}"
    plugin_dir.mkdir(parents=True, exist_ok=True)

    plugin_text = "\n".join(
        [
            "def list_all_profile_ids_v1():",
            f"    return {profile_ids!r}",
            "",
            "def list_needed_quirks_v1(profile_id):",
            f"    return {quirks_by_profile or {}!r}.get(profile_id, [])",
            "",
        ]
    )
    (plugin_dir / "mod.star").write_text(plugin_text, encoding="utf-8")


class TestCompositeRepoSingleEntry:
    def test_iter_repos_single(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        entry = _make_entry(local_path=str(tmp_path))
        composite = CompositeRepo([entry], gc)

        repos = list(composite.iter_repos())
        assert len(repos) == 1
        assert repos[0].repo_id == "test-repo"

    def test_inactive_entries_excluded(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        active_entry = _make_entry(repo_id="active", local_path=str(tmp_path))
        inactive_entry = _make_entry(
            repo_id="inactive", active=False, local_path=str(tmp_path / "other")
        )
        composite = CompositeRepo([active_entry, inactive_entry], gc)

        repos = list(composite.iter_repos())
        assert len(repos) == 1
        assert repos[0].repo_id == "active"

    def test_repo_id_returns_highest_priority(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        entry = _make_entry(repo_id="my-repo", priority=50, local_path=str(tmp_path))
        composite = CompositeRepo([entry], gc)

        assert composite.repo_id == "my-repo"

    def test_entries_sorted_by_priority(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        low = _make_entry(repo_id="low", priority=0, local_path=str(tmp_path / "low"))
        high = _make_entry(
            repo_id="high", priority=100, local_path=str(tmp_path / "high")
        )
        composite = CompositeRepo([high, low], gc)

        repos = list(composite.iter_repos())
        assert len(repos) == 2
        # Should be sorted ascending by priority
        assert repos[0].repo_id == "low"
        assert repos[1].repo_id == "high"


class TestCompositeRepoMultiEntryMerge:
    """Tests for multi-repo merge, deduplication, and priority shadowing."""

    def _make_composite(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> CompositeRepo:
        """Set up two repos with overlapping and unique packages.

        base (priority=0):
          toolchain/gcc 13.1.0 (slug=gcc-13)
          toolchain/gcc 13.2.0
          toolchain/llvm 17.0.0

        overlay (priority=100):
          toolchain/gcc 13.2.0 (shadows base, different desc)
          toolchain/gcc 14.0.0 (unique to overlay)
          source/some-lib 1.0.0 (unique category)
        """
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)

        base_root = tmp_path / "base"
        _init_repo_dir(base_root)
        _write_manifest(base_root, "toolchain", "gcc", "13.1.0", slug="gcc-13")
        _write_manifest(base_root, "toolchain", "gcc", "13.2.0", desc="base gcc 13.2")
        _write_manifest(base_root, "toolchain", "llvm", "17.0.0")

        overlay_root = tmp_path / "overlay"
        _init_repo_dir(overlay_root)
        _write_manifest(
            overlay_root, "toolchain", "gcc", "13.2.0", desc="overlay gcc 13.2"
        )
        _write_manifest(overlay_root, "toolchain", "gcc", "14.0.0")
        _write_manifest(overlay_root, "source", "some-lib", "1.0.0", slug="some-lib")

        base_entry = _make_entry(repo_id="base", priority=0, local_path=str(base_root))
        overlay_entry = _make_entry(
            repo_id="overlay", priority=100, local_path=str(overlay_root)
        )

        return CompositeRepo([overlay_entry, base_entry], gc)

    def test_iter_pkgs_merges_versions(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)
        pkgs = {(cat, name): vers for cat, name, vers in composite.iter_pkgs()}

        # gcc should have 3 versions merged
        assert ("toolchain", "gcc") in pkgs
        gcc_vers = pkgs[("toolchain", "gcc")]
        assert set(gcc_vers.keys()) == {"13.1.0", "13.2.0", "14.0.0"}

        # llvm from base only
        assert ("toolchain", "llvm") in pkgs
        assert set(pkgs[("toolchain", "llvm")].keys()) == {"17.0.0"}

        # some-lib from overlay only
        assert ("source", "some-lib") in pkgs

    def test_priority_shadowing_same_version(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        """When base and overlay both have gcc 13.2.0, overlay wins."""
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        pm = composite.get_pkg("gcc", "toolchain", "13.2.0")
        assert pm is not None
        # The overlay version should shadow the base.
        assert pm.repo_id == "overlay"
        assert pm.desc == "overlay gcc 13.2"

    def test_get_pkg_from_base_only(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        pm = composite.get_pkg("gcc", "toolchain", "13.1.0")
        assert pm is not None
        assert pm.repo_id == "base"

    def test_get_pkg_nonexistent(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)
        assert composite.get_pkg("nonexistent", "toolchain", "1.0.0") is None

    def test_get_pkg_latest_ver(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        latest = composite.get_pkg_latest_ver("gcc", "toolchain")
        assert latest.ver == "14.0.0"
        assert latest.repo_id == "overlay"

    def test_get_pkg_latest_ver_across_repos(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        """get_pkg_latest_ver without category uses merged by-name index."""
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        latest = composite.get_pkg_latest_ver("gcc")
        assert latest.ver == "14.0.0"

    def test_get_pkg_by_slug(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        # Slug from base repo
        pm = composite.get_pkg_by_slug("gcc-13")
        assert pm is not None
        assert pm.name == "gcc"

        # Slug from overlay repo
        pm = composite.get_pkg_by_slug("some-lib")
        assert pm is not None
        assert pm.name == "some-lib"

        assert composite.get_pkg_by_slug("nonexistent") is None

    def test_iter_pkg_vers_by_category(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        vers = list(composite.iter_pkg_vers("gcc", "toolchain"))
        ver_strs = {pm.ver for pm in vers}
        assert ver_strs == {"13.1.0", "13.2.0", "14.0.0"}

    def test_iter_pkg_vers_by_name_only(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        vers = list(composite.iter_pkg_vers("llvm"))
        assert len(vers) == 1
        assert vers[0].ver == "17.0.0"

    def test_iter_pkg_manifests_deduplicates(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        """iter_pkg_manifests should yield deduplicated manifests."""
        composite = self._make_composite(tmp_path, mock_gm, ruyi_logger)

        manifests = list(composite.iter_pkg_manifests())
        # Total unique (category, name, ver) tuples:
        # gcc: 13.1.0, 13.2.0, 14.0.0 = 3
        # llvm: 17.0.0 = 1
        # some-lib: 1.0.0 = 1
        # Total = 5
        keys = [(pm.category, pm.name, pm.ver) for pm in manifests]
        assert len(keys) == 5
        assert len(set(keys)) == 5  # all unique

    def test_slug_shadowing(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        """When two repos define the same slug, higher priority wins."""
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)

        base_root = tmp_path / "slug-base"
        _init_repo_dir(base_root)
        _write_manifest(
            base_root,
            "toolchain",
            "pkg-a",
            "1.0.0",
            slug="shared-slug",
            desc="from base",
        )

        overlay_root = tmp_path / "slug-overlay"
        _init_repo_dir(overlay_root)
        _write_manifest(
            overlay_root,
            "toolchain",
            "pkg-b",
            "2.0.0",
            slug="shared-slug",
            desc="from overlay",
        )

        base_entry = _make_entry(repo_id="base", priority=0, local_path=str(base_root))
        overlay_entry = _make_entry(
            repo_id="overlay", priority=100, local_path=str(overlay_root)
        )
        composite = CompositeRepo([base_entry, overlay_entry], gc)

        pm = composite.get_pkg_by_slug("shared-slug")
        assert pm is not None
        assert pm.desc == "from overlay"
        assert pm.repo_id == "overlay"

    def test_profile_queries_skip_broken_repo(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)

        base_root = tmp_path / "profile-base"
        _init_repo_dir(base_root)
        _write_profile_plugin(
            base_root,
            "riscv64",
            ["milkv-duo"],
            quirks_by_profile={"milkv-duo": ["xthead"]},
        )

        broken_root = tmp_path / "profile-broken"
        _init_repo_dir(broken_root)
        (broken_root / "plugins" / "ruyi-profile-riscv64").mkdir(
            parents=True,
            exist_ok=True,
        )

        base_entry = _make_entry(repo_id="base", priority=0, local_path=str(base_root))
        broken_entry = _make_entry(
            repo_id="broken",
            priority=100,
            local_path=str(broken_root),
        )
        composite = CompositeRepo([broken_entry, base_entry], gc)

        assert composite.get_supported_arches() == ["riscv64"]

        profiles = list(composite.iter_profiles_for_arch("riscv64"))
        assert [p.id for p in profiles] == ["milkv-duo"]
        assert profiles[0].need_quirks == {"xthead"}

        profile = composite.get_profile("milkv-duo")
        assert profile is not None
        assert profile.id == "milkv-duo"

        arch_profile = composite.get_profile_for_arch("riscv64", "milkv-duo")
        assert arch_profile is not None
        assert arch_profile.id == "milkv-duo"

    def test_profile_plugin_first_valid_repo_wins_per_arch(
        self,
        tmp_path: pathlib.Path,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)

        low_root = tmp_path / "profile-low"
        _init_repo_dir(low_root)
        _write_profile_plugin(
            low_root,
            "riscv64",
            ["low-profile"],
        )

        high_root = tmp_path / "profile-high"
        _init_repo_dir(high_root)
        _write_profile_plugin(
            high_root,
            "riscv64",
            ["high-profile"],
            quirks_by_profile={"high-profile": ["xthead"]},
        )

        low_entry = _make_entry(repo_id="low", priority=0, local_path=str(low_root))
        high_entry = _make_entry(
            repo_id="high",
            priority=100,
            local_path=str(high_root),
        )
        composite = CompositeRepo([high_entry, low_entry], gc)

        assert composite.get_supported_arches() == ["riscv64"]

        profiles = list(composite.iter_profiles_for_arch("riscv64"))
        assert [p.id for p in profiles] == ["high-profile"]
        assert profiles[0].need_quirks == {"xthead"}

        profile = composite.get_profile("high-profile")
        assert profile is not None
        assert profile.id == "high-profile"

        assert composite.get_profile("low-profile") is None
        assert composite.get_profile_for_arch("riscv64", "low-profile") is None



import os


class NewsReadStatusStore:
    def __init__(self, path: str) -> None:
        self._path = path
        self._status: set[str] = set()
        self._orig_status: set[str] = set()

    def load(self) -> None:
        try:
            with open(self._path, "r", encoding="utf-8") as fp:
                for line in fp:
                    self._orig_status.add(line.strip())
        except FileNotFoundError:
            return

        self._status = self._orig_status.copy()

    def __contains__(self, key: str) -> bool:
        return key in self._status

    def add(self, id: str) -> None:
        return self._status.add(id)

    def save(self) -> None:
        if self._status == self._orig_status:
            return

        content = "".join(f"{id}\n" for id in self._status)
        with open(self._path, "w", encoding="utf-8") as fp:
            fp.write(content)

    def remove(self) -> None:
        try:
            os.unlink(self._path)
        except FileNotFoundError:
            # nothing to remove, that's fine
            pass



"""
helper functions for CLI completions
"""

import argparse
from typing import Protocol, Any, TYPE_CHECKING

if TYPE_CHECKING:
    # A "lie" for type checking purposes. This is a known and wont fix issue for mypy.
    # Mypy would think the fallback import needs to be the same type as the first import.
    # See: https://github.com/python/mypy/issues/1153
    from argcomplete.completers import BaseCompleter
else:
    try:
        from argcomplete.completers import BaseCompleter
    except ImportError:
        # Fallback for environments where argcomplete is less than 2.0.0
        class BaseCompleter(object):
            def __call__(
                self,
                *,
                prefix: str,
                action: argparse.Action,
                parser: argparse.ArgumentParser,
                parsed_args: argparse.Namespace,
            ) -> None:
                raise NotImplementedError(
                    "This method should be implemented by a subclass."
                )


class NoneCompleter(BaseCompleter):
    def __call__(
        self,
        *,
        prefix: str,
        action: argparse.Action,
        parser: argparse.ArgumentParser,
        parsed_args: argparse.Namespace,
    ) -> None:
        return None


class DynamicCompleter(Protocol):
    def __call__(
        self,
        prefix: str,
        parsed_args: object,
        **kwargs: Any,
    ) -> list[str]: ...



from os import PathLike
from typing import (
    Any,
    Iterable,
    Mapping,
    Protocol,
    Sequence,
    TypedDict,
    TypeGuard,
    TYPE_CHECKING,
    cast,
)

if TYPE_CHECKING:
    from typing_extensions import NotRequired

from ..pluginhost.ctx import PluginHostContext
from ..pluginhost.traits import SupportsEvalFunction
from .entity_provider import BaseEntityProvider
from .pkg_manifest import EmulatorFlavor


class InvalidProfilePluginError(RuntimeError):
    def __init__(self, s: str) -> None:
        super().__init__(f"invalid arch profile plugin: {s}")


def validate_list_str(x: object) -> TypeGuard[list[str]]:
    if not isinstance(x, list):
        return False
    x = cast(list[object], x)
    return all(isinstance(y, str) for y in x)


def validate_list_str_or_none(x: object) -> TypeGuard[list[str] | None]:
    return True if x is None else validate_list_str(x)


def validate_dict_str_str(x: object) -> TypeGuard[dict[str, str]]:
    if not isinstance(x, dict):
        return False
    for k, v in cast(dict[object, object], x).items():
        if not isinstance(k, str) or not isinstance(v, str):
            return False
    return True


class PluginProfileProvider:
    def __init__(
        self,
        phctx: PluginHostContext[Any, SupportsEvalFunction],
        plugin_id: str,
    ) -> None:
        self._phctx = phctx
        self._plugin_id = plugin_id
        self._ev = phctx.make_evaluator()

    def _must_get(self, name: str) -> object:
        if v := self._phctx.get_from_plugin(self._plugin_id, name):
            return v
        raise InvalidProfilePluginError(
            f"'{name}' not found in plugin '{self._plugin_id}'"
        )

    def list_all_profile_ids(self) -> list[str]:
        fn = self._must_get("list_all_profile_ids_v1")
        ret = self._ev.eval_function(fn)
        if not validate_list_str(ret):
            raise InvalidProfilePluginError(
                "list_all_profile_ids must return list[str]"
            )

        return ret

    def list_needed_quirks(self, profile_id: str) -> list[str] | None:
        # For backward compatibility, try "list_needed_quirks_v1" first, then
        # fall back to "list_needed_flavors_v1" if the former is not available.
        fn = self._phctx.get_from_plugin(self._plugin_id, "list_needed_quirks_v1")
        if fn is None:
            fn = self._must_get("list_needed_flavors_v1")

        ret = self._ev.eval_function(fn, profile_id)
        if not validate_list_str_or_none(ret):
            raise InvalidProfilePluginError(
                "list_needed_quirks_v1 must return list[str] | None"
            )

        return ret

    def get_common_flags(self, profile_id: str, toolchain_quirks: list[str]) -> str:
        result = self._maybe_get_common_flags_v2(profile_id, toolchain_quirks)
        if result is not None:
            return result
        return self._get_common_flags_v1(profile_id)

    def _get_common_flags_v1(self, profile_id: str) -> str:
        fn = self._must_get("get_common_flags_v1")
        ret = self._ev.eval_function(fn, profile_id)
        if not isinstance(ret, str):
            raise InvalidProfilePluginError("get_common_flags_v1 must return str")

        return ret

    def _maybe_get_common_flags_v2(
        self,
        profile_id: str,
        toolchain_flavors: list[str],
    ) -> str | None:
        fn = self._phctx.get_from_plugin(self._plugin_id, "get_common_flags_v2")
        if fn is None:
            return None

        ret = self._ev.eval_function(fn, profile_id, toolchain_flavors)
        if not isinstance(ret, str):
            raise InvalidProfilePluginError("get_common_flags_v2 must return str")

        return ret

    def get_needed_emulator_pkg_flavors(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
    ) -> Iterable[str]:
        fn = self._must_get("get_needed_emulator_pkg_flavors_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
        )
        if not validate_list_str(ret):
            raise InvalidProfilePluginError(
                "get_needed_emulator_pkg_flavors_v1 must return list[str]"
            )

        return ret

    def check_emulator_flavor(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
        emulator_pkg_flavors: list[str] | None,
    ) -> bool:
        fn = self._must_get("check_emulator_flavor_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
            emulator_pkg_flavors,
        )
        if not isinstance(ret, bool):
            raise InvalidProfilePluginError("check_emulator_flavor_v1 must return bool")

        return ret

    def get_env_config_for_emu_flavor(
        self,
        profile_id: str,
        flavor: EmulatorFlavor,
        sysroot: PathLike[Any] | None,
    ) -> dict[str, str] | None:
        fn = self._must_get("get_env_config_for_emu_flavor_v1")
        ret = self._ev.eval_function(
            fn,
            profile_id,
            flavor,
            str(sysroot) if sysroot is not None else None,
        )
        if not validate_dict_str_str(ret):
            raise InvalidProfilePluginError(
                "get_env_config_for_emu_flavor_v1 must return dict[str, str]"
            )

        return ret


class ProfileProxy:
    def __init__(
        self,
        provider: PluginProfileProvider,
        arch: str,
        profile_id: str,
    ) -> None:
        self._provider = provider
        self._arch = arch
        self._id = profile_id

    @property
    def arch(self) -> str:
        return self._arch

    @property
    def id(self) -> str:
        return self._id

    @property
    def need_quirks(self) -> set[str]:
        r = self._provider.list_needed_quirks(self._id)
        return set(r) if r else set()

    def get_common_flags(self, toolchain_flavors: list[str]) -> str:
        return self._provider.get_common_flags(self._id, toolchain_flavors)

    def get_needed_emulator_pkg_flavors(
        self,
        flavor: EmulatorFlavor,
    ) -> set[str]:
        return set(self._provider.get_needed_emulator_pkg_flavors(self._id, flavor))

    def check_emulator_flavor(
        self,
        flavor: EmulatorFlavor,
        emulator_pkg_flavors: list[str] | None,
    ) -> bool:
        return self._provider.check_emulator_flavor(
            self._id, flavor, emulator_pkg_flavors
        )

    def get_env_config_for_emu_flavor(
        self,
        flavor: EmulatorFlavor,
        sysroot: PathLike[Any] | None,
    ) -> dict[str, str] | None:
        return self._provider.get_env_config_for_emu_flavor(self._id, flavor, sysroot)


#
# Protocols
#


# MetadataRepo is defined in repo.py, but we don't want to import repo.py here
# to avoid circular import. Instead, we just describe the methods and properties
# that we need from MetadataRepo with a Protocol.
class ProvidesProfiles(Protocol):
    def get_supported_arches(self) -> list[str]: ...
    def get_profile_for_arch(self, arch: str, name: str) -> ProfileProxy | None: ...
    def iter_profiles_for_arch(self, arch: str) -> Iterable[ProfileProxy]: ...


#
# Entity type and schema for profile entities
#

PROFILE_V1_ENTITY_TYPE = "profile-v1"
PROFILE_V1_ENTITY_TYPE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "required": ["profile-v1"],
    "properties": {
        "profile-v1": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "display_name": {"type": "string"},
                "name": {"type": "string"},
                "arch": {"type": "string"},
                "needed_toolchain_quirks": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "toolchain_common_flags_str": {"type": "string"},
            },
            "required": [
                "id",
                "display_name",
                "name",
                "arch",
                "needed_toolchain_quirks",
                "toolchain_common_flags_str",
            ],
        },
        "related": {
            "type": "array",
            "description": "List of related entity references",
            "items": {"type": "string", "pattern": "^.+:.+"},
        },
        "unique_among_type_during_traversal": {
            "type": "boolean",
            "description": "Whether this entity should be unique among all entities of the same type during traversal",
        },
    },
}


class ProfileV1EntityData(TypedDict):
    id: str
    display_name: str
    name: str
    arch: str
    needed_toolchain_quirks: list[str]
    toolchain_common_flags_str: str


ProfileV1Entity = TypedDict(
    "ProfileV1Entity",
    {
        "profile-v1": ProfileV1EntityData,
        "related": "NotRequired[list[str]]",
        "unique_among_type_during_traversal": "NotRequired[bool]",
    },
    total=False,
)


class ProfileEntityProvider(BaseEntityProvider):
    def __init__(self, provider: ProvidesProfiles) -> None:
        super().__init__()
        self._provider = provider

    def discover_schemas(self) -> dict[str, object]:
        return {
            PROFILE_V1_ENTITY_TYPE: PROFILE_V1_ENTITY_TYPE_SCHEMA,
        }

    def load_entities(
        self,
        entity_types: Sequence[str],
    ) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
        result: dict[str, Mapping[str, Mapping[str, Any]]] = {}
        for ty in entity_types:
            if ty == PROFILE_V1_ENTITY_TYPE:
                result[ty] = _load_profile_v1_entities(self._provider)
        return result


def _load_profile_v1_entities(provider: ProvidesProfiles) -> dict[str, ProfileV1Entity]:
    result: dict[str, ProfileV1Entity] = {}
    for arch in provider.get_supported_arches():
        result.update(_load_profile_v1_entities_for_arch(provider, arch))
    return result


def _load_profile_v1_entities_for_arch(
    provider: ProvidesProfiles,
    arch: str,
) -> dict[str, ProfileV1Entity]:
    result: dict[str, ProfileV1Entity] = {}
    for profile in provider.iter_profiles_for_arch(arch):
        full_name = profile.id
        relations = [f"arch:{arch}"]

        needed_toolchain_quirks = sorted(profile.need_quirks)

        result[profile.id] = {
            "profile-v1": {
                "id": profile.id,
                "display_name": full_name,
                "name": profile.id,
                "arch": profile.arch,
                "needed_toolchain_quirks": needed_toolchain_quirks,
                "toolchain_common_flags_str": profile.get_common_flags(
                    needed_toolchain_quirks,
                ),
            },
            "related": relations,
        }
    return result



import pathlib

from .canonical_dump import dumps_canonical_package_manifest_toml
from .pkg_manifest import PackageManifest


def load_package_manifest_from_path(path: pathlib.Path) -> PackageManifest:
    return PackageManifest.load_from_path(path)


def dump_canonical_package_manifest_from_path(path: pathlib.Path) -> str:
    return dumps_canonical_package_manifest_toml(
        load_package_manifest_from_path(path),
    )



import argparse
import os
import pathlib
import shutil
from typing import Final, TYPE_CHECKING

from ..i18n import _, d_
from .cmd import RootCommand

if TYPE_CHECKING:
    from .completion import ArgumentParser
    from .. import config


UNINSTALL_NOTICE: Final = d_(
    """
[bold]Thanks for hacking with [yellow]Ruyi[/]![/]

This will uninstall [yellow]Ruyi[/] from your system, and optionally remove
all installed packages and [yellow]Ruyi[/]-managed repository data if the
[green]--purge[/] switch is given on the command line.

Note that your [yellow]Ruyi[/] virtual environments [bold]will become unusable[/] after
[yellow]Ruyi[/] is uninstalled. You should take care of migrating or cleaning
them yourselves afterwards.
"""
)


# Self-management commands
class SelfCommand(
    RootCommand,
    cmd="self",
    has_subcommands=True,
    help=_("Manage this Ruyi installation"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        pass


class SelfCleanCommand(
    SelfCommand,
    cmd="clean",
    help=_("Remove various Ruyi-managed data to reclaim storage"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "--quiet",
            "-q",
            action="store_true",
            help=_("Do not print out the actions being performed"),
        )
        p.add_argument(
            "--all",
            action="store_true",
            help=_("Remove all covered data"),
        )
        p.add_argument(
            "--distfiles",
            action="store_true",
            help=_("Remove all downloaded distfiles if any"),
        )
        p.add_argument(
            "--installed-pkgs",
            action="store_true",
            help=_("Remove all installed packages if any"),
        )
        p.add_argument(
            "--news-read-status",
            action="store_true",
            help=_("Mark all news items as unread"),
        )
        p.add_argument(
            "--progcache",
            action="store_true",
            help=_("Clear the Ruyi program cache"),
        )
        p.add_argument(
            "--repo",
            action="store_true",
            help=_("Remove the Ruyi repo if located in Ruyi-managed cache directory"),
        )
        p.add_argument(
            "--telemetry",
            action="store_true",
            help=_("Remove all telemetry data recorded if any"),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        logger = cfg.logger
        quiet: bool = args.quiet
        all: bool = args.all
        distfiles: bool = args.distfiles
        installed_pkgs: bool = args.installed_pkgs
        news_read_status: bool = args.news_read_status
        progcache: bool = args.progcache
        repo: bool = args.repo
        telemetry: bool = args.telemetry

        if all:
            distfiles = True
            installed_pkgs = True
            news_read_status = True
            progcache = True
            repo = True
            telemetry = True

        if not any(
            [
                distfiles,
                installed_pkgs,
                news_read_status,
                progcache,
                repo,
                telemetry,
            ]
        ):
            logger.F(_("no data specified for cleaning"))
            logger.I(
                _(
                    "please check [yellow]ruyi self clean --help[/] for a list of cleanable data"
                )
            )
            return 1

        _do_reset(
            cfg,
            quiet=quiet,
            # state-related
            all_state=all,
            news_read_status=news_read_status,
            telemetry=telemetry,
            # cache-related
            all_cache=all,
            distfiles=distfiles,
            installed_pkgs=installed_pkgs,
            progcache=progcache,
            repo=repo,
        )

        return 0


class SelfUninstallCommand(
    SelfCommand,
    cmd="uninstall",
    help=_("Uninstall Ruyi"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "config.GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        p.add_argument(
            "--purge",
            action="store_true",
            help=_("Remove all installed packages and Ruyi-managed remote repo data"),
        )
        p.add_argument(
            "-y",
            action="store_true",
            dest="consent",
            help=_(
                "Give consent for uninstallation on CLI; do not ask for confirmation"
            ),
        )

    @classmethod
    def main(cls, cfg: "config.GlobalConfig", args: argparse.Namespace) -> int:
        from . import user_input

        logger = cfg.logger
        purge: bool = args.purge
        consent: bool = args.consent
        logger.D(f"ruyi self uninstall: purge={purge}, consent={consent}")

        if cfg.is_installation_externally_managed:
            logger.F(
                _(
                    "this [yellow]ruyi[/] is externally managed, for example, by the system package manager, and cannot be uninstalled this way"
                )
            )
            logger.I(_("please uninstall via the external manager instead"))
            return 1

        if not cfg.is_packaged:
            logger.F(
                _(
                    "this [yellow]ruyi[/] is not in standalone form, and cannot be uninstalled this way"
                )
            )
            return 1

        if not consent:
            logger.stdout(_(UNINSTALL_NOTICE))
            if not user_input.ask_for_yesno_confirmation(logger, _("Continue?")):
                logger.I(_("aborting uninstallation"))
                return 0
        else:
            logger.I(_("uninstallation consent given over CLI, proceeding"))

        _do_reset(
            cfg,
            quiet=False,
            installed_pkgs=purge,
            all_state=purge,
            all_cache=purge,
            self_binary=True,
        )

        logger.I(_("[yellow]ruyi[/] is uninstalled"))

        return 0


def _do_reset(
    cfg: "config.GlobalConfig",
    quiet: bool = False,
    *,
    installed_pkgs: bool = False,
    all_state: bool = False,
    news_read_status: bool = False,  # ignored if all_state=True
    telemetry: bool = False,  # ignored if all_state=True
    all_cache: bool = False,
    distfiles: bool = False,  # ignored if all_cache=True
    progcache: bool = False,  # ignored if all_cache=True
    repo: bool = False,  # ignored if all_cache=True
    self_binary: bool = False,
) -> None:
    logger = cfg.logger

    def status(s: str) -> None:
        if quiet:
            return
        logger.I(s)

    if installed_pkgs:
        status(_("removing installed packages"))
        shutil.rmtree(cfg.data_root, True)
        cfg.ruyipkg_global_state.purge_installation_info()

    # do not record any telemetry data if we're purging it
    if all_state or telemetry:
        cfg.telemetry.discard_events(True)

    if all_state:
        status(_("removing state data"))
        shutil.rmtree(cfg.state_root, True)
    else:
        if news_read_status:
            status(_("removing read status of news items"))
            cfg.news_read_status.remove()

        if telemetry:
            status(_("removing all telemetry data"))
            shutil.rmtree(cfg.telemetry_root, True)

    if all_cache:
        status(_("removing cached data"))
        shutil.rmtree(cfg.cache_root, True)
    else:
        if distfiles:
            status(_("removing downloaded distfiles"))
            # TODO: deduplicate the path derivation
            shutil.rmtree(os.path.join(cfg.cache_root, "distfiles"), True)

        if progcache:
            status(_("clearing the Ruyi program cache"))
            # TODO: deduplicate the path derivation
            shutil.rmtree(os.path.join(cfg.cache_root, "progcache"), True)

        if repo:
            # for safety, don't remove the repo if it's outside of Ruyi's XDG
            # cache root
            repo_dir = pathlib.Path(cfg.get_repo_dir()).resolve()
            cache_root = pathlib.Path(cfg.cache_root).resolve()

            repo_is_below_cache_root = False
            for p in repo_dir.parents:
                if p == cache_root:
                    repo_is_below_cache_root = True
                    break

            if not repo_is_below_cache_root:
                logger.W(
                    _(
                        "not removing the Ruyi repo: it is outside of the Ruyi cache directory"
                    )
                )
            else:
                status(_("removing the Ruyi repo"))
                shutil.rmtree(repo_dir, True)

    if self_binary:
        status(_("removing the ruyi binary"))
        try:
            os.unlink(cfg.self_exe)
        except FileNotFoundError:
            # we might have already removed ourselves during the purge; nothing to
            # do now.
            pass



import atexit
import os
import re
import shlex
from typing import Final, List, NoReturn

from ..config import GlobalConfig
from ..i18n import _
from ..utils.global_mode import ProvidesGlobalMode
from .venv_cfg import RuyiVenvConfig


def _run_exit_handlers_and_execv(
    path: str,
    argv: list[str],
) -> NoReturn:
    # run all exit handlers before execv
    # crucially this includes our telemetry handler so we don't lose telemetry
    # events in mux mode
    atexit._run_exitfuncs()
    os.execv(path, argv)


def mux_main(
    gm: ProvidesGlobalMode,
    gc: GlobalConfig,
    argv: List[str],
) -> int | NoReturn:
    basename = os.path.basename(gm.argv0)
    logger = gc.logger
    logger.D(f"mux mode: argv = {argv}, basename = {basename}")

    vcfg = RuyiVenvConfig.load_from_venv(gm, logger)
    if vcfg is None:
        logger.F(_("the Ruyi toolchain mux is not configured"))
        logger.I(_("check out `ruyi venv` for making a virtual environment"))
        return 1

    direct_symlink_target = resolve_direct_symlink_target(gm.argv0, vcfg)
    if direct_symlink_target is not None:
        logger.D(
            f"detected direct symlink target: {direct_symlink_target}, overriding basename"
        )
        basename = direct_symlink_target

    if basename == "ruyi-qemu":
        return mux_qemu_main(gc, vcfg, argv)

    # match the basename with one of the configured target tuples
    target_tuple: str | None = None
    binpath: str | None = None
    toolchain_sysroot: str | None = None
    toolchain_flags: str | None = None
    gcc_install_dir: str | None = None

    # prefer v1+ cached info which is lossless
    if md := vcfg.resolve_cmd_metadata_with_cache(basename):
        target_tuple = md["target_tuple"]
        binpath = md["dest"]
        if target_tuple:
            tgt_data = vcfg.targets.get(target_tuple)
            if tgt_data is None:
                logger.F(
                    _(
                        "internal error: no target data for tuple [yellow]{target_tuple}[/]"
                    ).format(
                        target_tuple=target_tuple,
                    )
                )
                return 1
            toolchain_sysroot = tgt_data.get("toolchain_sysroot")
            toolchain_flags = tgt_data.get("toolchain_flags")
            gcc_install_dir = tgt_data.get("gcc_install_dir")
    else:
        toolchain_bindir: str | None = None
        for tgt_tuple, tgt_data in vcfg.targets.items():
            if not basename.startswith(f"{tgt_tuple}-"):
                continue

            logger.D(f"matched target '{tgt_tuple}', data {tgt_data}")
            target_tuple = tgt_tuple
            toolchain_bindir = tgt_data["toolchain_bindir"]
            toolchain_sysroot = tgt_data.get("toolchain_sysroot")
            toolchain_flags = tgt_data.get("toolchain_flags")
            gcc_install_dir = tgt_data.get("gcc_install_dir")
            break

        if toolchain_bindir is None:
            # should not happen
            logger.F(
                _(
                    "internal error: no bindir configured for target [yellow]{target_tuple}[/]"
                ).format(
                    target_tuple=target_tuple,
                )
            )
            return 1

        binpath = os.path.join(toolchain_bindir, basename)

    if target_tuple is None:
        logger.F(
            _("no configured target found for command [yellow]{basename}[/]").format(
                basename=basename,
            )
        )
        return 1

    logger.D(f"binary to exec: {binpath}")

    argv_to_insert: list[str] | None = None
    if is_proxying_to_cc(basename):
        logger.D(f"{basename} is considered a CC")

        argv_to_insert = []

        if is_proxying_to_clang(basename):
            logger.D(f"adding target for clang: {target_tuple}")
            argv_to_insert.append(f"--target={target_tuple}")
            if gcc_install_dir is not None:
                logger.D(f"informing clang of GCC install dir: {gcc_install_dir}")
                argv_to_insert.append(f"--gcc-install-dir={gcc_install_dir}")

        if toolchain_flags is not None:
            argv_to_insert.extend(shlex.split(toolchain_flags))
            logger.D(f"parsed toolchain flags: {argv_to_insert}")

        if toolchain_sysroot is not None:
            logger.D(f"adding sysroot: {toolchain_sysroot}")
            argv_to_insert.extend(("--sysroot", toolchain_sysroot))

    new_argv = [binpath]
    if argv_to_insert:
        new_argv.extend(argv_to_insert)
    if len(argv) > 1:
        new_argv.extend(argv[1:])

    ensure_venv_in_path(vcfg)

    logger.D(f"exec-ing with argv {new_argv}")
    return _run_exit_handlers_and_execv(binpath, new_argv)


# TODO: dedup with venv provision logic (into a command name parser)
CC_ARGV0_RE: Final = re.compile(
    r"(?:^|-)(?:g?cc|c\+\+|g\+\+|cpp|clang|clang\+\+|clang-cl|clang-cpp)(?:-[0-9.]+)?$"
)


def resolve_direct_symlink_target(argv0: str, vcfg: RuyiVenvConfig) -> str | None:
    direct_symlink_target = resolve_argv0_symlink(argv0, vcfg)
    if direct_symlink_target is not None and os.path.sep in direct_symlink_target:
        # we're not designed to handle such indirections
        return None
    return direct_symlink_target


def resolve_argv0_symlink(argv0: str, vcfg: RuyiVenvConfig) -> str | None:
    if os.path.sep in argv0:
        # argv[0] contains path information that we can just use
        try:
            return os.readlink(argv0)
        except OSError:
            # argv[0] is not a symlink
            return None

    # argv[0] is bare command name, in which case we expect venv root to
    # be available, so we can just check f'{venv_root}/bin/{argv[0]}'.
    # we're guaranteed a venv_root because of the vcfg init logic.
    try:
        return os.readlink(vcfg.venv_root / "bin" / argv0)
    except OSError:
        return None


def is_proxying_to_cc(argv0: str) -> bool:
    return CC_ARGV0_RE.search(argv0) is not None


def is_proxying_to_clang(basename: str) -> bool:
    return "clang" in basename


def mux_qemu_main(
    gc: GlobalConfig,
    vcfg: RuyiVenvConfig,
    argv: List[str],
) -> int | NoReturn:
    logger = gc.logger
    binpath = vcfg.qemu_bin
    if binpath is None:
        logger.F(_("this virtual environment has no QEMU-like emulator configured"))
        return 1

    if vcfg.profile_emu_env is not None:
        logger.D(f"seeding QEMU environment with {vcfg.profile_emu_env}")
        for k, v in vcfg.profile_emu_env.items():
            os.environ[k] = v

    logger.D(f"QEMU binary to exec: {binpath}")

    new_argv = [binpath]
    if len(argv) > 1:
        new_argv.extend(argv[1:])

    logger.D(f"exec-ing with argv {new_argv}")
    return _run_exit_handlers_and_execv(binpath, new_argv)


def ensure_venv_in_path(vcfg: RuyiVenvConfig) -> None:
    venv_root = vcfg.venv_root
    venv_bindir = venv_root / "bin"
    venv_bindir = venv_bindir.resolve()

    orig_path = os.environ.get("PATH", "")
    for p in orig_path.split(os.pathsep):
        try:
            if os.path.samefile(p, venv_bindir):
                # TODO: what if our bindir actually comes after the system ones?
                return
        except FileNotFoundError:
            # maybe the PATH entry is stale
            continue

    # we're not in PATH, so prepend the bindir to PATH
    os.environ["PATH"] = f"{venv_bindir}:{orig_path}" if orig_path else str(venv_bindir)



import pathlib

import pytest

from ruyi.pluginhost.paths import resolve_ruyi_load_path


@pytest.fixture
def plugin_root(tmp_path: pathlib.Path) -> pathlib.Path:
    root = tmp_path / "plugins"
    (root / "alpha").mkdir(parents=True)
    (root / "alpha" / "mod.star").write_text("")
    (root / "alpha" / "sub.star").write_text("")
    (root / "alpha" / "data").mkdir()
    (root / "alpha" / "data" / "foo.toml").write_text("")
    (root / "beta").mkdir()
    (root / "beta" / "mod.star").write_text("")
    (root / "beta" / "data").mkdir()
    (root / "beta" / "data" / "bar.toml").write_text("")
    return root


def test_plain_relative_within_plugin(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    resolved = resolve_ruyi_load_path(
        "sub.star", plugin_root, False, originating, False
    )
    assert resolved == (plugin_root / "alpha" / "sub.star").resolve()


def test_plain_absolute_resolves_against_plugin_root(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    resolved = resolve_ruyi_load_path(
        "/sub.star", plugin_root, False, originating, False
    )
    assert resolved == plugin_root / "alpha" / "sub.star"


def test_plain_cross_plugin_boundary_rejected(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(ValueError, match="cross plugin boundary"):
        resolve_ruyi_load_path(
            "../beta/mod.star", plugin_root, False, originating, False
        )


def test_ruyi_plugin_scheme_resolves_entrypoint(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    resolved = resolve_ruyi_load_path(
        "ruyi-plugin://beta", plugin_root, False, originating, False
    )
    assert resolved == plugin_root / "beta" / "mod.star"


def test_ruyi_plugin_scheme_rejects_data_context(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="ruyi-plugin protocol"):
        resolve_ruyi_load_path(
            "ruyi-plugin://beta", plugin_root, True, originating, False
        )


def test_ruyi_plugin_scheme_rejects_path_segment(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="non-empty path segment"):
        resolve_ruyi_load_path(
            "ruyi-plugin://beta/extra", plugin_root, False, originating, False
        )


def test_ruyi_plugin_scheme_rejects_empty_netloc(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="empty location"):
        resolve_ruyi_load_path("ruyi-plugin://", plugin_root, False, originating, False)


def test_ruyi_plugin_data_scheme_resolves_under_data(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    resolved = resolve_ruyi_load_path(
        "ruyi-plugin-data://beta/bar.toml",
        plugin_root,
        True,
        originating,
        False,
    )
    assert resolved == plugin_root / "beta" / "data" / "bar.toml"


def test_ruyi_plugin_data_scheme_rejects_non_data_context(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="ruyi-plugin-data protocol"):
        resolve_ruyi_load_path(
            "ruyi-plugin-data://beta/bar.toml",
            plugin_root,
            False,
            originating,
            False,
        )


def test_host_scheme_requires_allow_host_fs_access(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="host protocol"):
        resolve_ruyi_load_path(
            "host:///etc/hostname", plugin_root, False, originating, False
        )


def test_host_scheme_returns_absolute_path_when_allowed(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    resolved = resolve_ruyi_load_path(
        "host:///etc/hostname", plugin_root, False, originating, True
    )
    assert resolved == pathlib.Path("/etc/hostname")


def test_fancy_uri_features_rejected(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="fancy URI features"):
        resolve_ruyi_load_path(
            "ruyi-plugin://beta?x=1", plugin_root, False, originating, False
        )


def test_unknown_scheme_rejected(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="unsupported Ruyi Starlark load path"):
        resolve_ruyi_load_path(
            "gopher://whatever", plugin_root, False, originating, False
        )


def test_double_slash_prefix_rejected(plugin_root: pathlib.Path) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="'//' is not allowed"):
        resolve_ruyi_load_path("//evil", plugin_root, False, originating, False)


# --- ruyi-build:// / ruyi-build-data:// ---------------------------------


@pytest.fixture
def recipe_project(tmp_path: pathlib.Path) -> pathlib.Path:
    root = tmp_path / "recipes_proj"
    (root / "lib").mkdir(parents=True)
    (root / "lib" / "docker.star").write_text("")
    (root / "cfg.toml").write_text("")
    (root / "recipes").mkdir()
    (root / "recipes" / "pkg.star").write_text("")
    return root


def test_ruyi_build_resolves_against_project_root(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    resolved = resolve_ruyi_load_path(
        "ruyi-build://lib/docker.star",
        plugin_root,
        False,
        originating,
        False,
        recipe_project_root=recipe_project,
    )
    assert resolved == (recipe_project / "lib" / "docker.star").resolve()


def test_ruyi_build_rejected_outside_recipe_context(
    plugin_root: pathlib.Path,
) -> None:
    originating = plugin_root / "alpha" / "mod.star"
    with pytest.raises(RuntimeError, match="only available when loading"):
        resolve_ruyi_load_path(
            "ruyi-build://lib/docker.star",
            plugin_root,
            False,
            originating,
            False,
        )


def test_ruyi_build_rejects_traversal(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    with pytest.raises(RuntimeError, match="escapes recipe project root"):
        resolve_ruyi_load_path(
            "ruyi-build://../../etc/passwd",
            plugin_root,
            False,
            originating,
            False,
            recipe_project_root=recipe_project,
        )


def test_ruyi_build_rejects_data_context(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    with pytest.raises(RuntimeError, match="ruyi-build protocol"):
        resolve_ruyi_load_path(
            "ruyi-build://cfg.toml",
            plugin_root,
            True,
            originating,
            False,
            recipe_project_root=recipe_project,
        )


def test_ruyi_build_data_resolves(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    resolved = resolve_ruyi_load_path(
        "ruyi-build-data://cfg.toml",
        plugin_root,
        True,
        originating,
        False,
        recipe_project_root=recipe_project,
    )
    assert resolved == (recipe_project / "cfg.toml").resolve()


def test_ruyi_build_data_rejects_non_data_context(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    with pytest.raises(RuntimeError, match="ruyi-build-data protocol"):
        resolve_ruyi_load_path(
            "ruyi-build-data://cfg.toml",
            plugin_root,
            False,
            originating,
            False,
            recipe_project_root=recipe_project,
        )


def test_ruyi_build_rejects_empty_path(
    plugin_root: pathlib.Path, recipe_project: pathlib.Path
) -> None:
    originating = recipe_project / "recipes" / "pkg.star"
    with pytest.raises(RuntimeError, match="empty path"):
        resolve_ruyi_load_path(
            "ruyi-build://",
            plugin_root,
            False,
            originating,
            False,
            recipe_project_root=recipe_project,
        )



#!/usr/bin/env python3

import os
import pathlib
import platform
import shutil
import subprocess
import sys
import time
import tomllib
from typing import Iterable, cast

from pygit2.repository import Repository
from rich.console import Console

try:
    from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
except ModuleNotFoundError:
    from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]

# it seems force_terminal is needed for colors to show up on GHA
INFO = Console(stderr=True, style="bold green", force_terminal=True, highlight=False)

LGPL_MODULES: Iterable[str] = ()


def make_canonicalized_host_for_progcache() -> str:
    # Similar to ruyipkg/host.py but in the format of f"{os}-{arch}"
    # (separated with a hyphen instead of slash)
    os = sys.platform
    if os == "win32":
        os = "windows"

    arch = platform.machine().lower()
    match arch:
        case "amd64" | "em64t":
            arch = "x86_64"
        case "arm64":
            arch = "aarch64"
        case "x86":
            arch = "i686"

    return f"{os}-{arch}"


def main() -> None:
    epoch = int(time.time())

    target_host = make_canonicalized_host_for_progcache()
    vers = get_versions()
    INFO.print(f"Target host              : [cyan]{target_host}")
    INFO.print(f"Project Git commit       : [cyan]{vers['git_commit']}")
    INFO.print(f"Project SemVer           : [cyan]{vers['semver']}")
    INFO.print(f"Version for use by Nuitka: [cyan]{vers['nuitka_ver']}")

    build_root = os.environ["RUYI_DIST_BUILD_DIR"]
    exe_name = "ruyi.exe" if sys.platform == "win32" else "ruyi"
    output_file = os.path.join(build_root, exe_name)

    cache_root = os.environ["RUYI_DIST_CACHE_DIR"]
    ensure_dir(cache_root)

    cache_key = get_cache_key(vers["git_commit"])
    cached_output_dir = pathlib.Path(cache_root) / cache_key
    cached_output_file = cached_output_dir / exe_name
    try:
        shutil.copyfile(cached_output_file, output_file)
        os.chmod(output_file, 0o755)
        INFO.print(f"cache hit at [cyan]{cached_output_file}[/], skipping build")
        return
    except FileNotFoundError:
        pass

    ext_outdir = os.path.join(build_root, "_exts")
    ensure_dir(ext_outdir)
    add_pythonpath(ext_outdir)

    # Compile LGPL module(s) into own extensions, if any
    if LGPL_MODULES:
        INFO.print("\nBuilding LGPL extension(s)\n")
        for name in LGPL_MODULES:
            make_nuitka_ext(name, ext_outdir)

    # Finally the main program
    INFO.print("\nBuilding Ruyi executable\n")
    begin_group("Building Ruyi executable")
    call_nuitka(
        "--standalone",
        "--onefile",
        "--assume-yes-for-downloads",
        "--output-filename=ruyi",
        f"--output-dir={build_root}",
        "--no-deployment-flag=self-execution",
        f"--product-version={vers['nuitka_ver']}",
        f"--onefile-tempdir-spec={{CACHE_DIR}}/ruyi/progcache/{vers['semver']}/{target_host}",
        "--include-package=charset_normalizer",
        "--include-package=pygments.formatters",
        "--include-package=pygments.lexers",
        "--include-package=pygments.styles",
        "--include-package=rich._unicode_data",
        "--include-package=_cffi_backend",  # https://github.com/Nuitka/Nuitka/issues/2505
        "--windows-icon-from-ico=resources/ruyi.ico",
        "--show-scons",
        "./ruyi/__main__.py",
    )
    end_group()

    begin_group("Cache maintenance")
    INFO.print(f"\ncaching output to [cyan]{cached_output_file}")
    ensure_dir(cached_output_dir)
    shutil.copyfile(output_file, cached_output_file)
    os.chmod(cached_output_file, 0o755)
    ts = cached_output_dir / "timestamp"
    ts.write_text(f"{epoch}\n")

    delete_cached_files_older_than_days(cache_root, 21, epoch)
    end_group()


def is_in_gha() -> bool:
    return "GITHUB_ACTIONS" in os.environ


def begin_group(title: str) -> None:
    if is_in_gha():
        print(f"::group::{title}", flush=True)


def end_group() -> None:
    if is_in_gha():
        print("::endgroup::", flush=True)


def ensure_dir(d: str | pathlib.Path) -> None:
    try:
        os.mkdir(d)
    except FileExistsError:
        pass


def get_cache_key(git_commit: str) -> str:
    return f"ruyi-g{git_commit}"


def delete_cached_files_older_than_days(root: str, days: int, epoch: int) -> None:
    max_ts_delta = days * 86400

    epoch_str = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(epoch))
    INFO.print(
        f"purging cache contents older than [cyan]{days}[/] days from [cyan]now={epoch_str}"
    )

    root_path = pathlib.Path(root)
    dirs_to_remove: list[tuple[pathlib.Path, int | None]] = []
    for f in root_path.iterdir():
        if f.name.startswith("pygit2"):
            INFO.print(f"ignoring library artifact cache [cyan]{f}")
            continue

        ts: int | None
        try:
            ts = int((f / "timestamp").read_text().strip(), 10)
        except (FileNotFoundError, ValueError):
            dirs_to_remove.append((f, None))
            continue

        if ts - epoch >= max_ts_delta:
            dirs_to_remove.append((f, ts))

    for f, ts in dirs_to_remove:
        if ts is None:
            INFO.print(
                f"removing [cyan]{f}[/] ([yellow]timestamp absent or invalid)[/]"
            )
        else:
            ts_time = time.gmtime(ts)
            ts_str = time.strftime("%Y-%m-%dT%H:%M:%SZ", ts_time)
            INFO.print(f"removing [cyan]{f}[/] (created [yellow]{ts_str}[/])")

        shutil.rmtree(f)


def call_nuitka(*args: str) -> None:
    nuitka_args = [
        # https://stackoverflow.com/questions/64761870/python-subprocess-doesnt-inherit-virtual-environment
        sys.executable,  # "python",
        "-m",
        "nuitka",
    ]
    nuitka_args.extend(args)
    subprocess.run(nuitka_args, check=True)


def add_pythonpath(path: str) -> None:
    old_path = os.environ.get("PYTHONPATH", "")
    new_path = path if not old_path else f"{path}{os.pathsep}{old_path}"
    os.environ["PYTHONPATH"] = new_path


def make_nuitka_ext(module_name: str, out_dir: str) -> None:
    mod = __import__(module_name)
    mod_dir = os.path.dirname(cast(str, mod.__file__))
    INFO.print(f"Building [cyan]{module_name}[/] at [cyan]{mod_dir}[/] into extension")
    begin_group(f"Building {module_name} into extension")
    call_nuitka(
        "--module",
        mod_dir,
        f"--include-package={module_name}",
        f"--output-dir={out_dir}",
    )
    end_group()


def get_versions() -> dict[str, str]:
    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)
    with open("pyproject.toml", "rb") as fp:
        pyproject = tomllib.load(fp)

    try:
        version = pyproject["project"]["version"]
    except KeyError:
        # In case the packaging environment has Poetry 1.x metadata switched
        # in
        version = pyproject["tool"]["poetry"]["version"]

    return {
        "git_commit": get_git_commit(),
        "semver": version,
        "nuitka_ver": to_version_for_nuitka(version),
    }


def get_git_commit() -> str:
    repo = Repository(".")
    return str(repo.head.target)


PRERELEASE_NUITKA_PATCH_VER_MAP = {
    "alpha": 10000,
    "beta": 20000,
    "rc": 30000,
}


def to_version_for_nuitka(version: str) -> str:
    """
    Figure out the Windows-style version string for Nuitka, from the input
    semver string.

    * `X.Y.Z` -> `X.Y.Z.0`
    * `X.Y.Z-alpha.YYYYMMDD` -> `X.(Y-1).1YYYY.MMDD0`
    * `X.Y.Z-beta.YYYYMMDD` -> `X.(Y-1).2YYYY.MMDD0`
    * `X.Y.Z-rc.YYYYMMDD` -> `X.(Y-1).3YYYY.MMDD0`

    The strange mapping is due to Nuitka (actually Windows?) requiring each
    part to fit in an u16.
    """

    sv = Version.parse(version)
    if not sv.prerelease:
        return f"{version}.0"

    n_major = sv.major
    n_minor = sv.minor - 1
    prerelease_kind, ymd_str = sv.prerelease.split(".")
    y, md = divmod(int(ymd_str), 10000)
    n_patch = PRERELEASE_NUITKA_PATCH_VER_MAP[prerelease_kind] + y
    n_extra = md * 10
    return f"{n_major}.{n_minor}.{n_patch}.{n_extra}"


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    main()



import platform
import sys
from typing import NamedTuple


class RuyiHost(NamedTuple):
    os: str
    arch: str

    def __str__(self) -> str:
        return f"{self.os}/{self.arch}"

    def canonicalize(self) -> "RuyiHost":
        return RuyiHost(
            os=canonicalize_os_str(self.os),
            arch=canonicalize_arch_str(self.arch),
        )


def canonicalize_host_str(host: str | RuyiHost) -> str:
    if isinstance(host, str):
        frags = host.split("/", 1)
        os = "linux" if len(frags) == 1 else frags[0]
        arch = frags[0] if len(frags) == 1 else frags[1]
        return str(RuyiHost(os, arch).canonicalize())

    return str(host.canonicalize())


def canonicalize_arch_str(arch: str) -> str:
    # Information sources:
    #
    # * https://bugs.python.org/issue7146#msg94134
    # * https://superuser.com/questions/305901/possible-values-of-processor-architecture
    match arch.lower():
        case "amd64" | "em64t":
            return "x86_64"
        case "arm64":
            return "aarch64"
        case "x86":
            return "i686"
        case arch_lower:
            return arch_lower


def canonicalize_os_str(os: str) -> str:
    match os:
        case "win32":
            return "windows"
        case _:
            return os


def get_native_host() -> RuyiHost:
    return RuyiHost(os=sys.platform, arch=platform.machine()).canonicalize()



import datetime
from functools import cached_property
import locale
import os.path
from os import PathLike
import pathlib
import sys
from typing import Any, Final, Iterable, Sequence, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import NotRequired, Self

    from ..log import RuyiLogger
    from ..ruyipkg.composite_repo import CompositeRepo
    from ..ruyipkg.repo import MetadataRepo, RepoEntry
    from ..ruyipkg.state import RuyipkgGlobalStateStore
    from ..telemetry.provider import TelemetryProvider
    from ..utils.global_mode import ProvidesGlobalMode
    from ..utils.xdg_basedir import XDGPathEntry
    from .news import NewsReadStatusStore

import babel

# not sure why Pyright insists on individual imports
# otherwise, at the use site (`except babel.core.UnknownLocaleError`):
#   error: "core" is not a known attribute of module "babel" (reportAttributeAccessIssue)
from babel.core import UnknownLocaleError

from ..i18n import _
from . import errors
from . import schema


if sys.platform == "linux":
    PRESET_GLOBAL_CONFIG_LOCATIONS: Final[list[str]] = [
        # TODO: enable distro packagers to customize the $PREFIX to suit their
        # particular FS layout if necessary.
        "/usr/share/ruyi/config.toml",
        "/usr/local/share/ruyi/config.toml",
    ]
else:
    PRESET_GLOBAL_CONFIG_LOCATIONS: Final[list[str]] = []

DEFAULT_APP_NAME: Final = "ruyi"
DEFAULT_REPO_URL: Final = "https://github.com/ruyisdk/packages-index.git"
DEFAULT_REPO_BRANCH: Final = "main"
DEFAULT_TELEMETRY_MODE: Final = "local"  # "off", "local", "on"


def get_host_path_fragment_for_binary_install_dir(canonicalized_host: str) -> str:
    # e.g. linux/amd64 -> amd64; "windows/amd64" -> "windows-amd64"
    if canonicalized_host.startswith("linux/"):
        return canonicalized_host[6:]
    return canonicalized_host.replace("/", "-")


def _get_lang_code() -> str:
    lang = locale.getlocale()[0]
    return lang or "en_US"


class GlobalConfigPackagesType(TypedDict):
    prereleases: "NotRequired[bool]"


class GlobalConfigRepoType(TypedDict):
    local: "NotRequired[str]"
    remote: "NotRequired[str]"
    branch: "NotRequired[str]"


class GlobalConfigInstallationType(TypedDict):
    # Undocumented: whether this Ruyi installation is externally managed.
    #
    # Can be used by distro packagers (by placing a config file in /etc/xdg/ruyi)
    # to signify this status to an official Ruyi build (where IS_PACKAGED is
    # True), to prevent e.g. accidental self-uninstallation.
    externally_managed: "NotRequired[bool]"


class GlobalConfigTelemetryType(TypedDict):
    mode: "NotRequired[str]"
    upload_consent: "NotRequired[datetime.datetime | str]"
    pm_telemetry_url: "NotRequired[str]"


class GlobalConfigReposEntryType(TypedDict):
    id: str
    name: "NotRequired[str]"
    remote: "NotRequired[str]"
    branch: "NotRequired[str]"
    local: "NotRequired[str]"
    priority: "NotRequired[int]"
    active: "NotRequired[bool]"


class GlobalConfigRootType(TypedDict):
    installation: "NotRequired[GlobalConfigInstallationType]"
    packages: "NotRequired[GlobalConfigPackagesType]"
    repo: "NotRequired[GlobalConfigRepoType]"
    repos: "NotRequired[list[GlobalConfigReposEntryType]]"
    telemetry: "NotRequired[GlobalConfigTelemetryType]"


class GlobalConfig:
    def __init__(self, gm: "ProvidesGlobalMode", logger: "RuyiLogger") -> None:
        from ..utils.xdg_basedir import XDGBaseDir

        self._gm = gm
        self.logger = logger

        # all defaults
        self.override_repo_dir: str | None = None
        self.override_repo_url: str | None = None
        self.override_repo_branch: str | None = None
        self.include_prereleases = False
        self.is_installation_externally_managed = False

        self._lang_code = _get_lang_code()

        self._dirs = XDGBaseDir(DEFAULT_APP_NAME)

        self._telemetry_mode: str | None = None
        self._telemetry_upload_consent: datetime.datetime | None = None
        self._telemetry_pm_telemetry_url: str | None = None

        self._extra_repo_entries: list["RepoEntry"] = []

    def _apply_config(
        self,
        config_data: GlobalConfigRootType,
        *,
        is_global_scope: bool,
    ) -> None:
        if ins_cfg := config_data.get(schema.SECTION_INSTALLATION):
            iem = ins_cfg.get(schema.KEY_INSTALLATION_EXTERNALLY_MANAGED, None)
            if iem is not None and not is_global_scope:
                iem_cfg_key = f"{schema.SECTION_INSTALLATION}.{schema.KEY_INSTALLATION_EXTERNALLY_MANAGED}"
                self.logger.W(
                    _(
                        "the config key [yellow]{key}[/] cannot be set from user config; ignoring"
                    ).format(
                        key=iem_cfg_key,
                    ),
                )
            else:
                self.is_installation_externally_managed = bool(iem)

        if pkgs_cfg := config_data.get(schema.SECTION_PACKAGES):
            self.include_prereleases = pkgs_cfg.get(
                schema.KEY_PACKAGES_PRERELEASES, False
            )

        if repo_cfg := config_data.get(schema.SECTION_REPO):
            self.override_repo_dir = repo_cfg.get(schema.KEY_REPO_LOCAL, None)
            self.override_repo_url = repo_cfg.get(schema.KEY_REPO_REMOTE, None)
            self.override_repo_branch = repo_cfg.get(schema.KEY_REPO_BRANCH, None)

            if self.override_repo_dir:
                if not pathlib.Path(self.override_repo_dir).is_absolute():
                    self.logger.W(
                        _(
                            "the local repo path '{path}' is not absolute; ignoring"
                        ).format(
                            path=self.override_repo_dir,
                        )
                    )
                    self.override_repo_dir = None

        if tele_cfg := config_data.get(schema.SECTION_TELEMETRY):
            self._telemetry_mode = tele_cfg.get(schema.KEY_TELEMETRY_MODE, None)
            self._telemetry_pm_telemetry_url = tele_cfg.get(
                schema.KEY_TELEMETRY_PM_TELEMETRY_URL,
                None,
            )

            self._telemetry_upload_consent = None
            if consent := tele_cfg.get(schema.KEY_TELEMETRY_UPLOAD_CONSENT, None):
                if isinstance(consent, datetime.datetime):
                    self._telemetry_upload_consent = consent

        if repos_cfg := config_data.get(schema.SECTION_REPOS):
            self._parse_repos_config(repos_cfg, is_system=is_global_scope)

    def _parse_repos_config(
        self,
        repos_cfg: "list[GlobalConfigReposEntryType]",
        *,
        is_system: bool = False,
    ) -> None:
        from ..ruyipkg.repo import (
            DEFAULT_REPO_ID,
            DEFAULT_REPO_PRIORITY,
            REPO_ID_PATTERN,
            RepoEntry,
        )

        seen_ids: set[str] = {e.id for e in self._extra_repo_entries}
        new_entries: list[RepoEntry] = []

        for entry_data in repos_cfg:
            repo_id = entry_data.get(schema.KEY_REPOS_ID, "")
            if not repo_id or not REPO_ID_PATTERN.match(repo_id):
                self.logger.W(
                    _("ignoring [[repos]] entry with invalid id: '{id}'").format(
                        id=repo_id
                    )
                )
                continue

            if repo_id == DEFAULT_REPO_ID:
                self.logger.W(
                    _(
                        "ignoring [[repos]] entry with reserved id '{id}'; "
                        "use [repo] to configure the default repository"
                    ).format(id=repo_id)
                )
                continue

            if repo_id in seen_ids:
                self.logger.W(
                    _("ignoring duplicate [[repos]] entry with id '{id}'").format(
                        id=repo_id
                    )
                )
                continue

            remote = entry_data.get(schema.KEY_REPOS_REMOTE, "")
            local_path = entry_data.get(schema.KEY_REPOS_LOCAL, None)
            if not remote and not local_path:
                self.logger.W(
                    _(
                        "ignoring [[repos]] entry '{id}': "
                        "at least one of 'remote' or 'local' must be set"
                    ).format(id=repo_id)
                )
                continue

            if local_path and not pathlib.Path(local_path).is_absolute():
                self.logger.W(
                    _(
                        "ignoring [[repos]] entry '{id}': "
                        "the local path '{path}' is not absolute"
                    ).format(id=repo_id, path=local_path)
                )
                continue

            seen_ids.add(repo_id)
            new_entries.append(
                RepoEntry(
                    id=repo_id,
                    name=entry_data.get(schema.KEY_REPOS_NAME, repo_id),
                    remote=remote,
                    branch=entry_data.get(schema.KEY_REPOS_BRANCH, DEFAULT_REPO_BRANCH),
                    local_path=local_path,
                    priority=entry_data.get(
                        schema.KEY_REPOS_PRIORITY, DEFAULT_REPO_PRIORITY
                    ),
                    active=entry_data.get(schema.KEY_REPOS_ACTIVE, True),
                    is_system=is_system,
                )
            )

        self._extra_repo_entries.extend(new_entries)

    def get_by_key(self, key: str | Sequence[str]) -> object:
        parsed_key = schema.parse_config_key(key)
        section, sel = parsed_key[0], parsed_key[1:]
        attr_name = self._get_attr_name_by_key(section, sel)
        if attr_name is None:
            raise errors.InvalidConfigKeyError(key)
        return getattr(self, attr_name)

    def set_by_key(self, key: str | Sequence[str], value: object) -> None:
        # We don't have to check for global-only keys here because this
        # method is only used for programmatic changes to the in-memory
        # config, not for loading from config files.
        parsed_key = schema.parse_config_key(key)
        section, sel = parsed_key[0], parsed_key[1:]
        attr_name = self._get_attr_name_by_key(section, sel)
        if attr_name is None:
            raise errors.InvalidConfigKeyError(key)
        schema.ensure_valid_config_kv(key, True, value)
        setattr(self, attr_name, value)

    @classmethod
    def _get_attr_name_by_key(cls, section: str, sel: list[str]) -> str | None:
        if section == schema.SECTION_INSTALLATION:
            return cls._get_section_installation(sel)
        elif section == schema.SECTION_PACKAGES:
            return cls._get_section_packages(sel)
        elif section == schema.SECTION_REPO:
            return cls._get_section_repo(sel)
        elif section == schema.SECTION_TELEMETRY:
            return cls._get_section_telemetry(sel)
        else:
            return None

    @classmethod
    def _get_section_installation(cls, selector: list[str]) -> str | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_INSTALLATION_EXTERNALLY_MANAGED:
            return "is_installation_externally_managed"
        else:
            return None

    @classmethod
    def _get_section_packages(cls, selector: list[str]) -> str | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_PACKAGES_PRERELEASES:
            return "include_prereleases"
        else:
            return None

    @classmethod
    def _get_section_repo(cls, selector: list[str]) -> str | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_REPO_BRANCH:
            return "override_repo_branch"
        elif leaf == schema.KEY_REPO_LOCAL:
            return "override_repo_dir"
        elif leaf == schema.KEY_REPO_REMOTE:
            return "override_repo_url"
        else:
            return None

    @classmethod
    def _get_section_telemetry(cls, selector: list[str]) -> str | None:
        if len(selector) != 1:
            return None
        leaf = selector[0]
        if leaf == schema.KEY_TELEMETRY_MODE:
            return "telemetry_mode"
        elif leaf == schema.KEY_TELEMETRY_PM_TELEMETRY_URL:
            return "override_pm_telemetry_url"
        elif leaf == schema.KEY_TELEMETRY_UPLOAD_CONSENT:
            return "telemetry_upload_consent_time"
        else:
            return None

    @property
    def argv0(self) -> str:
        return self._gm.argv0

    @property
    def main_file(self) -> str:
        return self._gm.main_file

    @property
    def self_exe(self) -> str:
        return self._gm.self_exe

    @property
    def is_debug(self) -> bool:
        return self._gm.is_debug

    @property
    def is_experimental(self) -> bool:
        return self._gm.is_experimental

    @property
    def is_packaged(self) -> bool:
        return self._gm.is_packaged

    @property
    def is_porcelain(self) -> bool:
        return self._gm.is_porcelain

    @property
    def is_telemetry_optout(self) -> bool:
        return self._gm.is_telemetry_optout

    @property
    def is_cli_autocomplete(self) -> bool:
        return self._gm.is_cli_autocomplete

    @property
    def venv_root(self) -> str | None:
        return self._gm.venv_root

    @property
    def lang_code(self) -> str:
        return self._lang_code

    @cached_property
    def babel_locale(self) -> babel.Locale:
        try:
            return babel.Locale.parse(self.lang_code)
        except UnknownLocaleError:
            # this can happen in case of unrecognized locale names, which
            # apparently falls back to "C"
            return babel.Locale.parse("en_US")

    @property
    def cache_root(self) -> os.PathLike[Any]:
        return self._dirs.app_cache

    @property
    def data_root(self) -> os.PathLike[Any]:
        return self._dirs.app_data

    @property
    def state_root(self) -> os.PathLike[Any]:
        return self._dirs.app_state

    @cached_property
    def news_read_status(self) -> "NewsReadStatusStore":
        from .news import NewsReadStatusStore

        filename = os.path.join(self.ensure_state_dir(), "news.read.txt")
        return NewsReadStatusStore(filename)

    @property
    def telemetry_root(self) -> os.PathLike[Any]:
        return pathlib.Path(self.ensure_state_dir()) / "telemetry"

    @cached_property
    def telemetry(self) -> "TelemetryProvider":
        from ..telemetry.provider import TelemetryProvider

        # for allowing minimal uploads when telemetry is off
        minimal_mode = self.telemetry_mode == "off"
        return TelemetryProvider(self, minimal_mode)

    @property
    def telemetry_mode(self) -> str:
        return self._telemetry_mode or DEFAULT_TELEMETRY_MODE

    @telemetry_mode.setter
    def telemetry_mode(self, mode: str) -> None:
        if mode not in ("off", "local", "on"):
            raise ValueError("telemetry mode must be one of: off, local, on")
        if self._gm.is_telemetry_optout and mode != "off":
            raise ValueError(
                "cannot enable telemetry when the environment variable opt-out is set"
            )
        self._telemetry_mode = mode

    @property
    def telemetry_upload_consent_time(self) -> datetime.datetime | None:
        return self._telemetry_upload_consent

    @telemetry_upload_consent_time.setter
    def telemetry_upload_consent_time(self, t: datetime.datetime | None) -> None:
        self._telemetry_upload_consent = t

    @property
    def override_pm_telemetry_url(self) -> str | None:
        return self._telemetry_pm_telemetry_url

    @override_pm_telemetry_url.setter
    def override_pm_telemetry_url(self, url: str | None) -> None:
        self._telemetry_pm_telemetry_url = url

    @cached_property
    def default_repo_dir(self) -> str:
        # Prefer the new repos/<id> layout if it exists, otherwise fall back
        # to the legacy packages-index/ path for pre-migration state.
        new_path = os.path.join(self.cache_root, "repos", "ruyisdk")
        if os.path.isdir(new_path):
            return new_path
        return os.path.join(self.cache_root, "packages-index")

    def get_repo_dir(self) -> str:
        import warnings

        warnings.warn(
            "get_repo_dir() is deprecated; use get_repo_dir_for_id(repo_id) instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.override_repo_dir or self.default_repo_dir

    def get_repo_dir_for_id(self, repo_id: str) -> str:
        """Return the local checkout path for a repo identified by ``repo_id``.

        If the matching ``RepoEntry`` has a ``local_path`` set, that path is
        returned directly; otherwise the default ``<cache>/repos/<id>/``
        location is used.
        """
        for entry in self.repo_entries:
            if entry.id == repo_id:
                return entry.resolve_root(self.cache_root)
        return os.path.join(self.cache_root, "repos", repo_id)

    @cached_property
    def have_overridden_repo_dir(self) -> bool:
        if not self.override_repo_dir:
            return False
        override_path = pathlib.Path(self.override_repo_dir)
        default_path = pathlib.Path(self.default_repo_dir)
        # we don't use samefile() here because the path may not exist
        return override_path.resolve() != default_path.resolve()

    def get_repo_url(self) -> str:
        return self.override_repo_url or DEFAULT_REPO_URL

    def get_repo_branch(self) -> str:
        return self.override_repo_branch or DEFAULT_REPO_BRANCH

    @cached_property
    def repo_entries(self) -> "list[RepoEntry]":
        from ..ruyipkg.repo import RepoEntry

        entries = [RepoEntry.from_legacy_config(self)]
        entries.extend(self._extra_repo_entries)
        entries.sort(key=lambda e: e.priority)
        return entries

    @cached_property
    def repo(self) -> "CompositeRepo":
        from ..ruyipkg.composite_repo import CompositeRepo

        self._ensure_repo_layout_migrated()
        return CompositeRepo(self.repo_entries, self)

    @cached_property
    def default_repo(self) -> "MetadataRepo":
        """Return the default (ruyisdk) MetadataRepo for code that truly
        needs a specific MetadataRepo instance.

        .. deprecated::
            Use ``self.repo`` (CompositeRepo) instead for multi-repo-aware
            access, or iterate ``self.repo_entries`` for specific repos.
        """
        import warnings

        warnings.warn(
            "default_repo is deprecated; use the CompositeRepo via .repo instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self._ensure_repo_layout_migrated()
        return self.repo_entries[0].make_metadata_repo(self)

    def _ensure_repo_layout_migrated(self) -> None:
        """Run the one-time migration from packages-index/ to repos/ruyisdk/
        if the user has not overridden the repo directory."""
        if self.override_repo_dir:
            return

        from ..ruyipkg.migration import migrate_repo_dir

        migrate_repo_dir(str(self.cache_root), self.logger)

    def ensure_distfiles_dir(self) -> str:
        path = pathlib.Path(self.ensure_cache_dir()) / "distfiles"
        path.mkdir(parents=True, exist_ok=True)
        return str(path)

    def global_binary_install_root(self, host: str, slug: str) -> str:
        host_path = get_host_path_fragment_for_binary_install_dir(host)
        path = pathlib.Path(self.ensure_data_dir()) / "binaries" / host_path / slug
        return str(path)

    def global_blob_install_root(self, slug: str) -> str:
        path = pathlib.Path(self.ensure_data_dir()) / "blobs" / slug
        return str(path)

    def lookup_binary_install_dir(self, host: str, slug: str) -> PathLike[Any] | None:
        host_path = get_host_path_fragment_for_binary_install_dir(host)
        for data_dir in self._dirs.app_data_dirs:
            p = data_dir.path / "binaries" / host_path / slug
            if p.exists():
                return p
        return None

    @property
    def ruyipkg_state_root(self) -> os.PathLike[Any]:
        return pathlib.Path(self.ensure_state_dir()) / "ruyipkg"

    @cached_property
    def ruyipkg_global_state(self) -> "RuyipkgGlobalStateStore":
        from ..ruyipkg.state import RuyipkgGlobalStateStore

        return RuyipkgGlobalStateStore(self.ruyipkg_state_root)

    def ensure_data_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_data
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_cache_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_cache
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_config_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_config
        p.mkdir(parents=True, exist_ok=True)
        return p

    def ensure_state_dir(self) -> os.PathLike[Any]:
        p = self._dirs.app_state
        p.mkdir(parents=True, exist_ok=True)
        return p

    def iter_preset_configs(self) -> "Iterable[XDGPathEntry]":
        """
        Yields possible Ruyi config files in all preset config path locations,
        sorted by precedence from lowest to highest (so that each file may be
        simply applied consecutively).
        """

        from ..utils.xdg_basedir import XDGPathEntry

        for path in PRESET_GLOBAL_CONFIG_LOCATIONS:
            yield XDGPathEntry(pathlib.Path(path), True)

    def iter_xdg_configs(self) -> "Iterable[XDGPathEntry]":
        """
        Yields possible Ruyi config files in all XDG config paths, sorted by precedence
        from lowest to highest (so that each file may be simply applied consecutively).
        """

        from ..utils.xdg_basedir import XDGPathEntry

        entries = list(self._dirs.app_config_dirs)
        for e in reversed(entries):
            yield XDGPathEntry(e.path / "config.toml", e.is_global)

    @property
    def local_user_config_file(self) -> pathlib.Path:
        return self._dirs.app_config / "config.toml"

    def _try_apply_config_file(
        self,
        path: os.PathLike[Any],
        *,
        is_global_scope: bool,
    ) -> None:
        import tomlkit

        try:
            with open(path, "rb") as fp:
                data: Any = tomlkit.load(fp)
        except FileNotFoundError:
            return

        self.logger.D(f"applying config: {data}, is_global_scope={is_global_scope}")
        self._apply_config(data, is_global_scope=is_global_scope)

    @classmethod
    def load_from_config(cls, gm: "ProvidesGlobalMode", logger: "RuyiLogger") -> "Self":
        obj = cls(gm, logger)

        for config_path, is_global in obj.iter_preset_configs():
            obj.logger.D(f"trying config file from preset location: {config_path}")
            obj._try_apply_config_file(config_path, is_global_scope=is_global)

        for config_path, is_global in obj.iter_xdg_configs():
            obj.logger.D(f"trying config file from XDG path: {config_path}")
            obj._try_apply_config_file(config_path, is_global_scope=is_global)

        # let environment variable take precedence
        if gm.is_telemetry_optout:
            obj._telemetry_mode = "off"
            obj._telemetry_upload_consent = None

        return obj



from contextlib import AbstractContextManager
import enum
import io
import json
import sys
from types import TracebackType
from typing import Protocol, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self


class PorcelainEntityType(enum.StrEnum):
    LogV1 = "log-v1"
    NewsItemV1 = "newsitem-v1"
    PkgListOutputV1 = "pkglistoutput-v1"
    EntityListOutputV1 = "entitylistoutput-v1"
    RepoEntryV1 = "repoentry-v1"
    CheckDiagnosticV1 = "checkdiagnostic-v1"


class PorcelainEntity(TypedDict):
    ty: PorcelainEntityType


class PorcelainBinarySink(Protocol):
    def flush(self) -> None: ...
    def write(self, b: bytes, /) -> int: ...


class PorcelainTextSink(Protocol):
    def flush(self) -> None: ...
    def write(self, s: str, /) -> int: ...


class PorcelainOutput(AbstractContextManager["PorcelainOutput"]):
    def __init__(
        self,
        *,
        binary_out: PorcelainBinarySink | None = None,
        text_out: PorcelainTextSink | None = None,
    ) -> None:
        self._txt_out: PorcelainTextSink | None
        self._bin_out: PorcelainBinarySink | None

        if binary_out is None and text_out is None:
            if isinstance(sys.stdout, io.TextIOWrapper):
                self._bin_out = sys.stdout.buffer
                self._txt_out = None
            else:
                self._bin_out = None
                self._txt_out = sys.stdout
            return

        if binary_out is not None and text_out is not None:
            raise ValueError("cannot specify both binary_out and text_out")

        self._txt_out = text_out
        self._bin_out = binary_out

    def __enter__(self) -> "Self":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        if self._txt_out is not None:
            self._txt_out.flush()
        if self._bin_out is not None:
            self._bin_out.flush()
        return None

    def emit(self, obj: PorcelainEntity) -> None:
        s = json.dumps(obj, ensure_ascii=False, separators=(",", ":"))
        if self._txt_out is not None:
            self._txt_out.write(s)
            self._txt_out.write("\n")
            return
        assert self._bin_out is not None
        self._bin_out.write(s.encode("utf-8"))
        self._bin_out.write(b"\n")
